#
# (C) Copyright 2003, 2004 Hewlett-Packard Development Company, L.P.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Tim Potter <tpot@hp.com>

"""Useful CIM constants."""

# CIMError error code constants

CIM_ERR_FAILED                       = 1  # A general error occurred
CIM_ERR_ACCESS_DENIED                = 2  # Resource not available
CIM_ERR_INVALID_NAMESPACE            = 3  # The target namespace does not exist
CIM_ERR_INVALID_PARAMETER            = 4  # Parameter value(s) invalid
CIM_ERR_INVALID_CLASS                = 5  # The specified Class does not exist
CIM_ERR_NOT_FOUND                    = 6  # Requested object could not be found
CIM_ERR_NOT_SUPPORTED                = 7  # Operation not supported
CIM_ERR_CLASS_HAS_CHILDREN           = 8  # Class has subclasses
CIM_ERR_CLASS_HAS_INSTANCES          = 9  # Class has instances
CIM_ERR_INVALID_SUPERCLASS           = 10 # Superclass does not exist
CIM_ERR_ALREADY_EXISTS               = 11 # Object already exists
CIM_ERR_NO_SUCH_PROPERTY             = 12 # Property does not exist
CIM_ERR_TYPE_MISMATCH                = 13 # Value incompatible with type
CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED = 14 # Query language not supported
CIM_ERR_INVALID_QUERY                = 15 # Query not valid
CIM_ERR_METHOD_NOT_AVAILABLE         = 16 # Extrinsic method not executed
CIM_ERR_METHOD_NOT_FOUND             = 17 # Extrinsic method does not exist

# Provider types

PROVIDERTYPE_CLASS       = 1
PROVIDERTYPE_INSTANCE    = 2
PROVIDERTYPE_ASSOCIATION = 3
PROVIDERTYPE_INDICATION  = 4
PROVIDERTYPE_METHOD      = 5
PROVIDERTYPE_CONSUMER    = 6            # Indication consumer
PROVIDERTYPE_QUERY       = 7



#
# (C) Copyright 2003-2005 Hewlett-Packard Development Company, L.P.
# (C) Copyright 2006-2007 Novell, Inc. 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Tim Potter <tpot@hp.com>
# Author: Martin Pool <mbp@hp.com>
# Author: Bart Whiteley <bwhiteley@suse.de>

'''
This module implements CIM operations over HTTP.

This module should not know anything about the fact that the data
being transferred is XML.  It is up to the caller to format the input
data and interpret the result.
'''

import sys, string, re, os, socket 
# ASX don't import pwd HERE anymore
#import pwd
from stat import S_ISSOCK
import cim_obj
from types import StringTypes

class Error(Exception):
    """This exception is raised when a transport error occurs."""
    pass

class AuthError(Error):
    """This exception is raised when an authentication error (401) occurs."""
    pass

def parse_url(url):
    """Return a tuple of (host, port, ssl) from the URL parameter.
    The returned port defaults to 5988 if not specified.  SSL supports
    defaults to False if not specified."""

    host = url                          # Defaults
    port = 5988
    ssl = False

    if re.match("https", url):          # Set SSL if specified
        ssl = True
        port = 5989

    m = re.search("^https?://", url)    # Eat protocol name
    if m:
        host = url[len(m.group(0)):]

    s = string.split(host, ":")         # Set port number
    if len(s) != 1:
        host = s[0]
        port = int(s[1])

    return host, port, ssl

def wbem_request(url, data, creds, headers = [], debug = 0, x509 = None,
                 verify_callback = None):
    """Send XML data over HTTP to the specified url. Return the
    response in XML.  Uses Python's build-in httplib.  x509 may be a
    dictionary containing the location of the SSL certificate and key
    files."""

    import httplib, base64, urllib

    class HTTPBaseConnection:
        def send(self, str):
            """ Same as httplib.HTTPConnection.send(), except we don't 
            check for sigpipe and close the connection.  If the connection 
            gets closed, getresponse() fails. 
            """

            if self.sock is None:
                if self.auto_open:
                    self.connect()
                else:
                    raise httplib.NotConnected()
            if self.debuglevel > 0:
                print "send:", repr(str)
            self.sock.sendall(str)

    class HTTPConnection(HTTPBaseConnection, httplib.HTTPConnection):
        def __init__(self, host, port=None, strict=None):
            httplib.HTTPConnection.__init__(self, host, port, strict)
    
    class HTTPSConnection(HTTPBaseConnection, httplib.HTTPSConnection):
        def __init__(self, host, port=None, key_file=None, cert_file=None, 
                     strict=None):
            httplib.HTTPSConnection.__init__(self, host, port, key_file, 
                                             cert_file, strict)
    
    class FileHTTPConnection(HTTPBaseConnection, httplib.HTTPConnection):
        def __init__(self, uds_path):
            httplib.HTTPConnection.__init__(self, 'localhost')
            self.uds_path = uds_path
        def connect(self):
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.sock.connect(self.uds_path)

    host, port, ssl = parse_url(url)

    key_file = None
    cert_file = None

    if ssl:

        if x509 is not None:
            cert_file = x509.get('cert_file')
            key_file = x509.get('key_file')

        if verify_callback is not None:
            try:
                from OpenSSL import SSL
                ctx = SSL.Context(SSL.SSLv3_METHOD)
                ctx.set_verify(SSL.VERIFY_PEER, verify_callback)
                # Add the key and certificate to the session
                if cert_file is not None and key_file is not None:
                  ctx.use_certificate_file(cert_file)
                  ctx.use_privatekey_file(key_file)
                s = SSL.Connection(ctx, socket.socket(socket.AF_INET,
                                                      socket.SOCK_STREAM))
                s.connect((host, port))
                s.do_handshake()
                s.shutdown()
                s.close()
            except socket.error, arg:
                raise Error("Socket error: %s" % (arg,))
            except socket.sslerror, arg:
                raise Error("SSL error: %s" % (arg,))

    numTries = 0
    localAuthHeader = None
    tryLimit = 5

    data = '<?xml version="1.0" encoding="utf-8" ?>\n' + data

    local = False
    if ssl:
        h = HTTPSConnection(host, port = port, key_file = key_file,
                                            cert_file = cert_file)
    else:
        if url.startswith('http'):
            h = HTTPConnection(host, port = port)
        else:
            if url.startswith('file:'):
                url = url[5:]
            try:
                s = os.stat(url)
                if S_ISSOCK(s.st_mode):
                    h = FileHTTPConnection(url)
                    local = True
                else:
                    raise Error('Invalid URL')
            except OSError:
                raise Error('Invalid URL')

    locallogin = None
    if host in ('localhost', '127.0.0.1'):
        local = True
    if local:
        uid = os.getuid()
        try:
            import pwd
            locallogin = pwd.getpwuid(uid)[0]
        except KeyError:
            locallogin = None
    while numTries < tryLimit:
        numTries = numTries + 1

        h.putrequest('POST', '/cimom')

        h.putheader('Content-type', 'application/xml; charset="utf-8"')
        # ASX in 2.7 second arg must be string
        h.putheader('Content-length', str(len(data)))
        if localAuthHeader is not None:
            h.putheader(*localAuthHeader)
        elif creds is not None: 
            h.putheader('Authorization', 'Basic %s' %
                    base64.encodestring('%s:%s' % (creds[0], creds[1])).replace('\n',''))
        elif locallogin is not None:
            h.putheader('PegasusAuthorization', 'Local "%s"' % locallogin)

        for hdr in headers:
            s = map(lambda x: string.strip(x), string.split(hdr, ":", 1))
            h.putheader(urllib.quote(s[0]), urllib.quote(s[1]))

        try:
            # See RFC 2616 section 8.2.2
            # An http server is allowed to send back an error (presumably 
            # a 401), and close the connection without reading the entire
            # request.  A server may do this to protect itself from a DoS
            # attack.  
            # 
            # If the server closes the connection during our h.send(), we 
            # will either get a socket exception 104 (TCP RESET), or a 
            # socket exception 32 (broken pipe).  In either case, thanks
            # to our fixed HTTPConnection classes, we'll still be able to 
            # retrieve the response so that we can read and respond to the 
            # authentication challenge. 
            h.endheaders()
            try:
                h.send(data)
            except socket.error, arg:
                if arg[0] != 104 and arg[0] != 32:
                    raise

            response = h.getresponse()
            body = response.read()
    
            if response.status != 200:
                if response.status == 401:
                    if numTries >= tryLimit:
                        raise AuthError(response.reason)
                    if not local:
                        raise AuthError(response.reason)
                    authChal = response.getheader('WWW-Authenticate', '')
                    if 'openwbem' in response.getheader('Server', ''):
                        if 'OWLocal' not in authChal:
                            localAuthHeader = ('Authorization', 
                                    'OWLocal uid="%d"' % uid)
                            continue
                        else:
                            try:
                                nonceIdx = authChal.index('nonce=')
                                nonceBegin = authChal.index('"', nonceIdx)
                                nonceEnd = authChal.index('"', nonceBegin+1)
                                nonce = authChal[nonceBegin+1:nonceEnd]
                                cookieIdx = authChal.index('cookiefile=')
                                cookieBegin = authChal.index('"', cookieIdx)
                                cookieEnd = authChal.index('"', cookieBegin+1)
                                cookieFile = authChal[cookieBegin+1:cookieEnd]
                                f = open(cookieFile, 'r')
                                cookie = f.read().strip()
                                f.close()
                                localAuthHeader = ('Authorization',
                                        'OWLocal nonce="%s", cookie="%s"' % \
                                                (nonce, cookie))
                                continue
                            except:
                                localAuthHeader = None
                                continue
                    elif 'Local' in authChal:
                        try:
                            beg = authChal.index('"') + 1
                            end = authChal.rindex('"')
                            if end > beg:
                                file = authChal[beg:end]
                                fo = open(file, 'r')
                                cookie = fo.read().strip()
                                fo.close()
                                localAuthHeader = ('PegasusAuthorization',
                                    'Local "%s:%s:%s"' % \
                                            (locallogin, file, cookie))
                                continue
                        except ValueError:
                            pass

                    raise AuthError(response.reason)
                if response.getheader('CIMError', None) is not None and \
                   response.getheader('PGErrorDetail', None) is not None:
                    import urllib
                    raise Error(
                        'CIMError: %s: %s' %
                        (response.getheader('CIMError'),
                         urllib.unquote(response.getheader('PGErrorDetail'))))
                raise Error('HTTP error: %s' % response.reason)
    
        except httplib.BadStatusLine, arg:
            raise Error("The web server returned a bad status line: '%s'" % arg)
        except socket.error, arg:
            raise Error("Socket error: %s" % (arg,))
        except socket.sslerror, arg:
            raise Error("SSL error: %s" % (arg,))

        break

    return body


def get_object_header(obj):
    """Return the HTTP header required to make a CIM operation request
    using the given object.  Return None if the object does not need
    to have a header."""

    # Local namespacepath

    if isinstance(obj, StringTypes):
        return 'CIMObject: %s' % obj

    # CIMLocalClassPath

    if isinstance(obj, cim_obj.CIMClassName):
        return 'CIMObject: %s:%s' % (obj.namespace, obj.classname)

    # CIMInstanceName with namespace

    if isinstance(obj, cim_obj.CIMInstanceName) and obj.namespace is not None:
        return 'CIMObject: %s' % obj

    raise TypeError('Don\'t know how to generate HTTP headers for %s' % obj)



#
# ghettovcb.py
#
# mksbackup ghettovcb frontent
#

import os, string, subprocess, fnmatch, platform, time
import posixpath, ftplib
import threading, random

from datetime import datetime, timedelta
import Queue

import paramiko
import scp


from archiver import *
import ftpd
import cron

GHETTO_VERSION=[ '2011_03_14_1', '2011_06_28_1' ] 

exit_code_re=re.compile('^\s*exit_code\s*=\s*([0-9]+)\s*$')

# ---------------------------------------------------------------------------
class MySCPClient(scp.SCPClient):
    """MySCPClient.put converts all '\r\n' into '\n'
       MySCPClient.get downloads the remote dir at same level as local dir
    """ 
    
    # -----------------------------------------------------------------------
    def _send_files(self, files):
        for name in files:
            basename = os.path.basename(name)
            (mode, size, mtime, atime) = self._read_stats(name)
            if self.preserve_times:
                self._send_time(mtime, atime)
            file_hdl = file(name, 'rb')
            buffer=file_hdl.read()
            file_hdl.close()
            buffer=buffer.replace('\r', '')
            size=len(buffer)
            self.channel.sendall('C%s %d %s\n' % (mode, size, basename))
            self._recv_confirm()
            file_pos = 0
            buff_size = self.buff_size
            chan = self.channel
            while file_pos < size:
                buf=buffer[file_pos:file_pos+buff_size]
                chan.sendall(buf)
                file_pos += len(buf) 
                if self.callback:
                    self.callback(file_pos, size)
            chan.sendall('\x00')

    # -----------------------------------------------------------------------
    def get(self, remote_path, local_path = '', recursive = False, preserve_times = False):
        self._recv_topdir = local_path or os.getcwd()
        return scp.SCPClient.get(self, remote_path, local_path, recursive, preserve_times)
    
    # -----------------------------------------------------------------------
    def _recv_pushd(self, cmd):
        parts = cmd.split()
        try:
            mode = int(parts[0], 8)
            path = os.path.join(self._recv_dir, parts[2])
        except:
            self.channel.send('\x01')
            raise scp.SCPException('Bad directory format')
        else:
            print 'ASX', self._recv_dir, self._recv_topdir
            if self._recv_dir==self._recv_topdir and (not os.path.exists(self._recv_dir) or os.path.isdir(self._recv_dir)):
                # use self._recv_dir instead of os.path.join(self._recv_dir, parts[2])
                path=self._recv_dir
                self._recv_topdir=''

        try:
            if not os.path.exists(path):
                os.mkdir(path, mode)
            elif os.path.isdir(path):
                os.chmod(path, mode)
            else:
                raise scp.SCPException('%s: Not a directory' % path)
            self._dirtimes[path] = (self._utime)
            self._utime = None
            self._recv_dir = path
        except (OSError, scp.SCPException), e:
            self.channel.send('\x01'+e.message)
            raise
        

# ---------------------------------------------------------------------------
def RunCommand(t, command, log=None):
    chan=t.open_session()
    exit_code=None
    chan.set_combine_stderr(True)
    cmd_line=command+' ; echo exit_code=$?'
    chan.exec_command(cmd_line)
    output=''
    while 1:
        try:
            x=chan.recv(1024)
            if len(x)==0:
                break
            output+=x
        except socket.timeout:
            break

    exit_code=chan.recv_exit_status()
    chan.close()
    
    # ESX(i) don't return any usable exit_code, use the one returned by  "echo exit_code=$?" 
    pos1=output.rfind('exit_code=')
    pos2=output.find('\n', pos1)
    exit_code=int(output[pos1+10:pos2])
    output=output[:pos1]

    if log:
        l=log.debug
        if exit_code!=0:
            l=log.warning
            
        if output or exit_code!=0:
            l('exit_code=%d command=%s', exit_code, cmd_line)
            for line in output.split('\n'):
                l('> %s', line)
        
    return exit_code, output
  

class Ghettovcb(Archiver):
    
    name='ghettovcb'
    exe='ghettovcb'

    types=dict( backup='backup', copy='copy', move='move',)
    variables=dict(vm='vm_name')


    ghetto_version_format1=re.compile(r'(?P<year>\d{4})_(?P<month>\d{2})_(?P<day>\d{2})')

    initiate_backup_re=re.compile(r'.*info: Initiate backup for (?P<name>.+)')
    backup_duration_re=re.compile(r'.*info: Backup Duration:')
    config_vm_backup_volume_re=re.compile(r'.*info:\s*CONFIG\s*-\s*VM_BACKUP_VOLUME\s*=\s*(?P<vm_backup_volume>.+)')
    config_disk_backup_format_re=re.compile(r'.*info:\s*CONFIG\s*-\s*DISK_BACKUP_FORMAT\s*=\s*(?P<disk_backup_format>.+)')
    config_vm_backup_dir_naming_convention_re=re.compile(r'.*info:\s*CONFIG\s*-\s*VM_BACKUP_DIR_NAMING_CONVENTION\s*=\s*(?P<vm_backup_dir_naming_convention>.+)')
    find_vmdk_re=re.compile(r'.*debug: findVMDK\(\) - Searching for VMDK: "(?P<vmdk>.+)" to backup')
    ftp_url_re=ftp_url_re

    # -----------------------------------------------------------------------
    def __init__(self):
        Archiver.__init__(self)
        
        self.host=None
        self.login=None
        self.password=None
        self.local=None
        self.remote_temp=None
        self.vm_list=[]
        self.vm_exclude=[]
        
        self.ghettovcb=None
        self.ghettovcb_bin=None
        self.ghettovcb_version='unknown'
        self.global_conf=None
        self.temp_target=None
        self.default_var=None

        self.ghettovcb_wait_nfs=None
        
        self.destination=None
        self.type='backup'
        
        self.scp_bin=None
        self.scp_args=None
        
        self.upload_status=''
        self.upload_status_code=None
        self.upload_thread=None
        self.upload_queue=Queue.Queue()
        self.uploaded_filenames=dict()
        self.download_script=None
        self.ftp_local=None
        self.ftp_local_default=None
        self.ftp_server=None
        self.ftp_thread=None  

    # -----------------------------------------------------------------------
    def read_vm_conf(self, filename, default={}):
        """load important variables from VM configuration backup script""" 
        values=default.copy()
        for line in open(filename, 'r'):
            for var in ('VM_BACKUP_VOLUME', 'ENABLE_COMPRESSION', 'LAST_MODIFIED_DATE', 'VER', 'VERSION', 'ENABLE_NON_PERSISTENT_NFS',):
                if line.startswith('printUsage() {'):
                    break
                if line.startswith('%s=' % var):
                    _k, v=line.split('=', 1)
                    values[var]=v.strip()
                    
        return values
    
    # -----------------------------------------------------------------------
    def load(self, job, manager):
        
        self.log=log=manager.log
        now=manager.now
        
        errors, warnings, extra={}, {}, ''

        try:
            self.host=job['host'].encode('ascii')
        except KeyError:
            errors['host']='option mandatory'
        except UnicodeEncodeError:
            errors['host']='invalid hostname or ip address'
            
        else:
            if not valid_ipRE.match(self.host):
                if not valid_hostnameRE.match(self.host):
                    errors['host']='invalid hostname or ip address'
                else:
                    try:
                        ip=socket.gethostbyname(self.host)
                    except socket.gaierror:
                        errors['host']='cannot resolve hostname'

        try:
            self.port=int(job.get('port', 22))
        except ValueError:
            errors['port']='must be an integer'
        else:
            if not (0<self.port and self.port<65535):
                errors['port']='must be an integer between 1 and 65535'

        for name in ('login', 'password', 'local'):
            value=job.get(name, None)
            if not value:
                errors[name]='option mandatory'
            else:
                setattr(self, name, value)
                if name=='password':
                    self.register_password(value)

        self.remote_temp=job.get('remote_temp', '/tmp')

        try:
            self.vm_list=quoted_string_list(job.get('vm_list', ''))
        except ValueError:
            errors['vm_list']='not a valid quoted string list'

        try:
            self.vm_exclude=quoted_string_list(job.get('vm_exclude', ''))
        except ValueError:
            errors['vm_exclude']='not a valid quoted string list'

        if job.get('target', None):
            self.target=job['target']

        self.global_conf=job.get('global_conf', None)

        self.script=job.get('script', 'ghettoVCB.sh')

        self.ftp_local=job.get('ftp_local', None)
        if self.ftp_local:
            if self.ftp_url_re.match(self.ftp_local):
                password=self.ftp_url_re.match(self.ftp_local).group('password')
                if password:
                    self.register_password(password)
            else:
                errors['ftp_local']='invalid url: ftp://[username[:password]@]host[:port]/rootdir'

        if not os.path.isdir(self.local):
            errors['local']='not a directory'
        else:
            # search ghettoVCB.sh whatever the char case
            self.ghettovcb=None
            list_dir=os.listdir(self.local)
            if self.script in list_dir:
                self.ghettovcb=self.script
            else:
                for filename in os.listdir(self.local):
                    if filename.lower()==self.script.lower():
                        self.ghettovcb=filename
                        break

            if not self.ghettovcb:
                errors['local']='file not found: %s' % (self.script, )

            # search the global_conf whatever the case
            if self.global_conf:
                if not self.global_conf in list_dir:
                    global_conf=None
                    for filename in list_dir:
                        if filename.lower()==self.global_conf.lower():
                            global_conf=filename
                            break
                    if not global_conf:
                        errors['global_conf']='file not found: %s' % (global_conf, )
                        self.global_conf=None
                    else:
                        self.global_conf=global_conf
            
            # read ghettoVCB.sh and extract default variables
            self.default_var={}
            if self.ghettovcb:
                self.default_var=self.read_vm_conf(os.path.join(self.local, self.ghettovcb))
                # check ghettoVCB version
                self.ghettovcb_version=self.default_var.get('LAST_MODIFIED_DATE','')+'_'+(self.default_var.get('VER','') or self.default_var.get('VERSION', ''))
                self.ghettovcb_wait_nfs=self.ghetto_version_format1.match(self.default_var.get('LAST_MODIFIED_DATE',''))!=None
                if self.ghettovcb_wait_nfs:
                    log.debug('this version of ghettovcb wait before to unmount the NFS share')
                
                if not self.ghettovcb_version in GHETTO_VERSION:
                    warnings['local']='This version of ghettoVCB.sh (%s) script has not been tested with MKSBackup, be warned ! Check the download page for more.' % (self.ghettovcb_version, )
                
            # read and extract global_conf
            if self.global_conf:
                self.default_var=self.read_vm_conf(os.path.join(self.local, self.global_conf), self.default_var)
            
        self.destination=job.get('destination', None)
        if self.destination:
                 
            self.destination=self.destination.replace('\n','')
            try:
                self.destination=Destinations(self.destination, self)
            except (DestinationSyntaxError, cron.CronException), e:
                errors['destination']='syntax error: %s' % (str(e), )

        if job.get('scp_bin', None):
            try:
                self.scp_args=quoted_string_list(job.get('scp_bin'))
            except ValueError:
                errors['scp_bin']='not a valid quoted string list'
            else:
                self.scp_bin=self.scp_args[0]
                self.scp_args[0]=os.path.basename(self.scp_args[0])
                if not os.path.isfile(self.scp_bin):
                    errors['scp_bin']='file not found'

        if not errors and self.destination and self.vm_list:
            # very limited check of target directory
            for vm in self.vm_list:
                type, target=self.destination.match(now, self.night_shift, variables=dict(vm=vm))
                # I cannot check more than dirname because when COMPRESSED is 1
                # target is a filename and not a directory 
                if type in ('move', 'copy') and not target.startswith('ftp://'):
                    dirname=os.path.dirname(target)
                    try:
                        os.makedirs(dirname)
                    except:
                        pass
                    if not os.path.isdir(dirname):
                        errors['destination']='directory not found: %s' % (dirname, )
                    
        if not errors:
            # upload and setup files on the VMWARE server

            t=paramiko.Transport((self.host, self.port))
            t.connect(username=self.login, password=self.password)
            
            scpclient=MySCPClient(t)
            scpclient.put(self.local, self.remote_temp, recursive=True)

            self.temp_target=posixpath.join(self.remote_temp, os.path.basename(self.local))
            self.ghettovcb_bin=posixpath.join(self.temp_target, self.ghettovcb)
            
            self.include_vm_file=posixpath.join(self.temp_target, 'include_vm')
            self.exclude_vm_file=posixpath.join(self.temp_target, 'exclude_vm')
            
            _exit_code, _output=RunCommand(t, 'chmod +x "%s"' % (self.ghettovcb_bin, ), log)

            self.cmd_line='cd "%s" ; ./"%s" -d debug -c "%s"' % (self.temp_target, self.ghettovcb, self.temp_target)
            
            if self.vm_list:
                self.cmd_line+=' -f "%s"' % (self.include_vm_file,)       
                for i, vm in enumerate(self.vm_list):
                    if i==0:
                        _exit_code, _output=RunCommand(t, 'echo "%s" > "%s"' % (vm, self.include_vm_file), log)
                    else:
                        _exit_code, _output=RunCommand(t, 'echo "%s" >> "%s"' % (vm, self.include_vm_file), log)
            else:
                self.cmd_line+=' -a'       
                if self.vm_exclude:
                    self.cmd_line+=' -e "%s"' % (self.exclude_vm_file,)       
                    for i, vm in enumerate(self.vm_exclude):
                        if i==0:
                            _exit_code, _output=RunCommand(t, 'echo "%s" > "%s"' % (vm, self.exclude_vm_file), log)
                        else:
                            _exit_code, _output=RunCommand(t, 'echo "%s" >> "%s"' % (vm, self.exclude_vm_file), log)
            if self.global_conf:
                self.cmd_line+=' -g "%s"' % (posixpath.join(self.temp_target, self.global_conf),)
            
            self.cmd_line+=' ; echo exit_code=$?' # ESXi don't return the exit_code   

            extra+='cmd_line=%s\n' % self.cmd_line
            t.close()
            
            if self.scp_bin:
                self.scp_args+=[ '-P', str(self.port), '-q' ]
                if sys.platform in ('win32', ):
                    # putty allow cmd line password
                    self.scp_args+=[ '-batch', '-pw', self.password ]
                extra+='scp_cmd=%s %s@%s:%%source%% %%destination%%\n' % (self.hide_password(' '.join(self.scp_args)), self.login, self.host, ) 
                
        return errors, warnings, extra

    # -----------------------------------------------------------------------
    def upload_queue_add_old(self, protocol, con, vm, source, destination):
        self.upload_queue.put((protocol, con, vm, source, destination))
        if not self.upload_thread:
            self.upload_thread=threading.Thread(target=self.uploader_thread, name='uploader')
            self.upload_thread.daemon=True
            self.upload_thread.start()
        
    # -----------------------------------------------------------------------
    def uploader_thread_old(self):
        current_vm=None
        self.upload_status=''
        if self.upload_status_code==None:
            self.upload_status_code='OK'
        while True:
            protocol, con, vm, source, destination=self.upload_queue.get(True)
            if vm!=current_vm:
                if current_vm!=None:
                    # just finished  current_vm
                    if status!='OK':
                        self.upload_status_code='ERR'
                    self.upload_status+='upload_status-%s=%s\r\n' % (current_vm, status)
                    self.upload_status+='upload_end-%s=%s\r\n' % (current_vm, time.ctime(end))
                    self.upload_status+='upload_end_epoch-%s=%d\r\n' % (current_vm, end)

                start=time.time()
                current_vm=vm
                if current_vm!=None:
                    # start a new vm
                    start=time.time()
                    status='OK'
                    self.upload_status+='upload_start-%s=%s\r\n' % (current_vm, time.ctime(start))
                    self.upload_status+='upload_start_epoch-%s=%d\r\n' % (current_vm, start)
                    self.upload_status+='upload_src-%s=%s\r\n' % (current_vm, source)
                    self.upload_status+='upload_dst-%s=%s\r\n' % (current_vm, destination)

            if protocol=='done':
                break
                                           
            if protocol=='ftp':
                ftp_line='ftpput -P %s -u "%s" -p "%s" %s "%s" "%s"' % (con['port'], con['username'], con['password'], con['host'], destination, source)
                self.log.info('start upload of "%s" to "%s"', source, destination)
                cmd_line=self.hide_password(ftp_line)
                self.log.debug('ftp command: %s', cmd_line)
                exit_code, output=RunCommand(self.t, ftp_line, log=None)
                if exit_code==0:
                    self.log.debug('file uploaded: %s', destination)
                else:
                    status='ERR'
                    self.log.error('upload failed %s (%s)', destination, output)
            elif protocol=='ssh':
                self.log.info('start download from "%s" to "%s"', source, destination)

                if not self.scp_bin:
                    # use paramiko
                    try:
                        scpclient=MySCPClient(self.t, buff_size=65536)
                        self.log.debug('scp %s %s', source, destination)
                        scpclient.get(source, os.path.dirname(destination))
                    except scp.SCPException:
                        self.log.error('download failed: %s (%s)', destination, output)
                        status='ERR'
                else:
                    # use local SCP
                    scp_args=self.scp_args[:]
                    scp_args+=[ '%s@%s:%s' % (self.login, self.host, source), destination]
                    scp_cmd=self.hide_password(' '.join(scp_args))
                    self.log.debug('scp_cmd=%s', scp_cmd)
                    if sys.platform in ('win32', ):
                        scp_args=param_encode(scp_args, self.manager.default_encoding)
                    scp_process=subprocess.Popen(scp_args, executable=self.scp_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    scp_out, scp_err=scp_process.communicate()
                    if scp_out:
                        self.log.debug('copying %s: %s', source, scp_out)
                    if scp_err:
                        self.log.debug('copying %s: %s', source, scp_err)

                    if scp_process.returncode!=0:
                        self.log.error('download failed: %s (%s)', destination, scp_out+scp_err)
                        status='ERR'

            end=time.time()
        self.upload_status='upload_status=%s\r\n%s' % (self.upload_status_code, self.upload_status)

    # -----------------------------------------------------------------------
    def start_ftp_server(self):
        if self.ftp_server:
            return True

        rndpass=reduce(lambda x, y:x+random.choice('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'), range(8), '')
        self.register_password(rndpass)

        if self.ftp_local:  
            ftp=self.ftp_url_re.match(self.ftp_local).groupdict()
        else:
            ftp=dict(host='localhost')

        for k, v in dict(username=rndpass, password=rndpass, directory='/', port='0').iteritems():
            if ftp.get(k)==None:
                ftp[k]=v
        
        if ftp['host']=='localhost':
            ftp['host']=socket.gethostbyname(socket.gethostname())
            host=''        
        else:
            host=socket.gethostbyname(ftp['host'])

        if sys.platform in ('win32', ):
            # convert posix directory into windows path 
            # /c/backup => C:\backup 
            root=ftp['directory']
            if re.match('^/[a-zA-Z]/', root):
                root=root[1]+':'+root[2:].replace('/', '\\')
            else:
                root=''
            ftp['directory']=root
        
        try:
            if ftp['directory']:
                os.makedirs(ftp['directory'])
        except:
            pass
        
        if not ftp['directory'] or not os.path.isdir(ftp['directory']):
            self.log.error('Cannot start built-in ftp server, wrong root directory: "%s"', ftp['directory'])            
            return False

        self.ftp_server=ftpd.FTPserver(addr=(host, int(ftp['port'])), root=ftp['directory'], user=ftp['username'], password=ftp['password']) # allowed=self.host
        self.ftp_thread=threading.Thread(target=self.ftp_server.serve_forever, name='ftpserver', )
        self.ftp_thread.daemon=True
        self.ftp_thread.start()
        
        ftp['port']=self.ftp_server.socket.getsockname()[1]
        self.ftp_local_default=ftp

        return True
         
    # -----------------------------------------------------------------------
    def upload_vm(self, current_vm, root_backup_dir, now, compressed):
        if not self.upload_thread:
            self.upload_thread=threading.Thread(target=self.uploader_thread, name='uploader')
            self.upload_thread.daemon=True
            self.upload_thread.start()
        
        self.upload_queue.put((current_vm, root_backup_dir, now, compressed))

    # -----------------------------------------------------------------------
    def uploader_thread(self):
        try:
            if not self._uploader_thread():
                self.upload_status_code='ERR'
        except Exception, e:
            self.upload_status_code='ERR'
            self.log.exception('unexpected error in uploader_thread: %s', e)
            
    # -----------------------------------------------------------------------
    def _uploader_thread(self):

        self.upload_status=''
        self.download_script=''
        
        if self.upload_status_code==None:
            self.upload_status_code='OK'
        while True:
            current_vm, root_backup_dir, now, compressed=self.upload_queue.get(True)
            
            if current_vm==None:
                break

            type, target=self.destination.match(now, self.night_shift, variables=dict(vm=current_vm))
            # print "ASX upload_vmdk vm=%s type=%s target=%s root_backup_dir=%s vmdk=%s" %(current_vm, type, target, root_backup_dir, vmdk)
            if not target or type not in ('move', 'copy'):
                continue
                
            ftp_match=self.ftp_url_re.match(target)

            if ftp_match:
                ftp=ftp_match.groupdict()
                if ftp['directory']==None:
                    ftp['directory']=''
                if ftp['port']!=None:
                    ftp['port']=int(ftp['port'])
                    
                if ftp['host']=='localhost':
                    if not self.ftp_server:
                        try:
                            res=self.start_ftp_server()
                        except Exception, e:
                            self.log.exception('cannot start built-in ftp server')
                            return False
                        else:
                            if not res:
                                return False
                            self.log.info('ftpserver listening on %s:%d', self.ftp_server.socket.getsockname()[0], self.ftp_server.socket.getsockname()[1])

                    ftp['host']=self.ftp_local_default['host']
                    for k in ('port', 'username', 'password'):
                        if ftp.get(k)==None:
                            ftp[k]=self.ftp_local_default[k]
                else:
                    if not ftp['port']:
                        ftp['port']=21
                    if not ftp['username']:
                        ftp['username']='anonymous'
                    if not ftp['password']:
                        ftp['password']='anonymous@'
                        
                    # try to create directories on the remote ftp server from 
                    # this host. If it fail the directory must already exist.   
                    try:
                        dir=ftp['directory']
                        if compressed:
                            dir=posixpath.dirname(dir)
                        ftp_pre_create_dir(ftp['host'], ftp['port'], ftp['username'], ftp['password'], dir)
                    except Exception, e:
                        self.log.exception('creating directory: %s', dir)
                        pass
                    else:
                        self.uploaded_filenames[dir]=None
                        
                target=ftp['directory']
                protocol, con='ftp', ftp

                if not self.download_script:
                    self.download_script+='#\n'
                    self.download_script+='# script generated by MKSBackup to download backup back to ESX(i) host\n'
                    self.download_script+='#\n'
                    self.download_script+='HOST="%s"\n' % (con['host'], )
                    self.download_script+='PORT=%d\n' % (con['port'], )
                    self.download_script+='USERNAME="%s"\n' % (con['username'], )
                    self.download_script+='PASSWORD="your_password"\n'
                    if compressed:
                        self.download_script+='ROOT_DIR="%s"\n' % (posixpath.dirname(root_backup_dir), )
                    else:
                        self.download_script+='ROOT_DIR="%s"\n' % (posixpath.dirname(posixpath.dirname(root_backup_dir)), )
                    
                self.download_script+='#\n'
                self.download_script+='# Download of VM: %s\n' % (current_vm, )
                self.download_script+='#\n'
                self.download_script+='DOWNLOAD_DIR="${ROOT_DIR}/download-%s"\n' % (current_vm, )
                self.download_script+='mkdir -p "${DOWNLOAD_DIR}"\n'
                
            else:
                # SSH
                # create the target directory, if don't exist
                try:
                    if compressed:
                        dir=os.path.dirname(target)
                    else:
                        dir=target
                    if not os.path.isdir(dir):
                        os.makedirs(dir)
                except Exception, e:
                    self.log.error('creating directory %s: %s', dir, e)
                    return False

                protocol, con='ssh', self.t
                
            if compressed:
                find_src=posixpath.join(root_backup_dir, compressed)
            else:
                find_src=root_backup_dir

            find_cmd='find "%s" -type f -exec stat -c "%%b:%%s:%%Y:%%Z:%%n" {} \;' % (find_src, )

            start=time.time()
            status='OK'
            self.upload_status+='upload_start-%s=%s\r\n' % (current_vm, time.ctime(start))
            self.upload_status+='upload_start_epoch-%s=%d\r\n' % (current_vm, start)
            self.upload_status+='upload_src-%s=%s\r\n' % (current_vm, find_src)
            self.upload_status+='upload_dst-%s=%s\r\n' % (current_vm, target)
            
            filenames=set()
            finished=False # ghettoVCB is still working on the current_vm
            wait=60
            while not finished:
                
#                print 'ASX WAIT', wait
                while wait>0 and not finished:
                    time.sleep(1)
                    wait-=1
                    finished=current_vm in self.finished_vm
                wait=60 # for next time
                                    
                # get epoch from ESX(i)
                exit_code, output=RunCommand(self.t, 'date +%s', log=None)
                if exit_code:
                    self.log.error('cannot get epoch: %r', output)
                    return False

                # print 'ASX finished', current_vm, finished
                esx_epoch=int(output.split('\n')[0])
                exit_code, output=RunCommand(self.t, find_cmd, log=None)
                if exit_code:
                    self.log.error('cannot get file list of %s: %r', find_src, output)
                    return False

#      2496:1302887189:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01.vmx
#1190330368:1302887225:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/4970f796-06563af5-553e-000e0c7 7dba/cos5-tiny 1-s001.vmdk
#     65536:1302887191:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/4970f796-06563af5-553e-000e0c7 7dba/cos5-tiny 1-s002.vmdk
#       528:1302887225:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/4970f796-06563af5-553e-000e0c7 7dba/cos5-tiny 1.vmdk
#2072182784:1302887332:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s001.vmdk
#2102722560:1302887444:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s002.vmdk
#2106982400:1302887560:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s003.vmdk
#2102001664:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s004.vmdk
#   4259840:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s005.vmdk
#       649:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01.vmdk
#        30:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/STATUS.ok

                files=dict()
                for line in output.split('\n'):
                    if not line:
                        continue
                    block_number, size, c_epoch, m_epoch, filename=line.split(':', 5)
                    if filename.startswith('.lck-'):
                        continue
                    
                    files[filename]=(max(int(c_epoch), int(m_epoch)), line)
                         
                for filename, (epoch, line) in files.iteritems():
                    if line in self.uploaded_filenames:
                        continue
                    if not finished:
                        if epoch+30>esx_epoch and current_vm not in self.finished_vm:
                            # file last modified < 30s ago, skip 
                            continue

                        match=re.match('.*-s(?P<number>\d{3}).vmdk$', filename)
                        if match:
                            number=int(match.group('number'))
                            previous_part=filename[:match.start('number')]+'%03d' % (number-1, )+filename[match.end('number'):]
                            if previous_part in files and files[previous_part][-1] not in self.uploaded_filenames:
                                # part of 2gbsparse but previous part not yet uploaded
                                continue

                    source=filename
                                
                    if compressed:
                        relfilename='%s.tar.gz' % (current_vm, )
                    else:
                        relfilename=source[len(root_backup_dir):]
                        relfilename=relfilename.lstrip('/')
                    
                    destination=posixpath.join(target, relfilename)
                    dest_dir=posixpath.dirname(destination)
                    if protocol=='ftp':
                            
                        if dest_dir not in self.uploaded_filenames:
                            self.download_script+='mkdir -p "${DOWNLOAD_DIR}/%s"\n' % (posixpath.dirname(relfilename), )
                            try:
                                ftp_pre_create_dir(ftp['host'], ftp['port'], ftp['username'], ftp['password'], dest_dir)
                            except Exception, e:
                                self.log.error('cannot create directory, try to create it yourself on ftp server: %s', dest_dir)
                            else:
                                self.uploaded_filenames[dest_dir]=None
                            
                        ftp_line='ftpput -P %s -u "%s" -p "%s" %s "%s" "%s"' % (con['port'], con['username'], con['password'], con['host'], destination, source)
                        self.download_script+='echo -n download "%s"\n' % (destination, )
                        self.download_script+='ftpget -P ${PORT} -u "${USERNAME}" -p "${PASSWORD}" "${HOST}" "${DOWNLOAD_DIR}/%s" "%s" && echo " OK" || echo " ERR"\n' % (relfilename, destination, )
                        self.log.info('start upload of "%s" to "%s"', source, destination)
                        cmd_line=self.hide_password(ftp_line)
                        self.log.debug('ftp command: %s', cmd_line)
                        exit_code, output=RunCommand(self.t, ftp_line, log=None)
                        if exit_code==0:
                            self.log.debug('file uploaded: %s', destination)
                            self.uploaded_filenames[line]=None
                        else:
                            if finished:
                                status='ERR'
                                self.log.error('upload failed %s (%s)', destination, output)
                                self.log.debug('epoch=%d file=%s', esx_epoch, line)
                            else:
                                self.log.info('upload failed, retry later %s (%s)', destination, output)
                                
                    elif protocol=='ssh':
                        dest_dir=os.path.join(target, posixpath.dirname(destination))
                        try:
                            if not os.path.isdir(dest_dir):
                                self.log.debug('create directory %s', dest_dir)
                                os.makedirs(dest_dir)
                        except Exception, e:
                            self.log.error('creating directory %s: %s', target, e)
                            return False
                        
                        self.log.info('start download from "%s" to "%s"', source, destination)
        
                        # escape space characters
                        scp_source=source.replace(' ', '\\ ')
                        if not self.scp_bin:
                            # use paramiko
                            try:
                                scpclient=MySCPClient(self.t, buff_size=65536)
                                self.log.debug('scp %s %s', scp_source, destination)
                                scpclient.get(source, destination)
                            except scp.SCPException, e:
                                if finished:
                                    status='ERR'
                                    self.log.error('download failed: %s (%s)', destination, e)
                                    self.log.debug('epoch=%d file=%s', esx_epoch, line)
                                else:
                                    self.log.info('download failed, retry later %s (%s)', destination, e)
                        
                        else:
                            # use local SCP
                            scp_args=self.scp_args[:]
                            scp_args+=[ '%s@%s:%s' % (self.login, self.host, scp_source), destination]
                            scp_cmd=self.hide_password(' '.join(scp_args))
                            self.log.debug('scp_cmd=%s', scp_cmd)
                            if sys.platform in ('win32', ):
                                scp_args=param_encode(scp_args, self.manager.default_encoding)
                            scp_process=subprocess.Popen(scp_args, executable=self.scp_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                            scp_out, scp_err=scp_process.communicate()
                            if scp_out:
                                self.log.debug('copying %s: %s', source, scp_out)
                            if scp_err:
                                self.log.debug('copying %s: %s', source, scp_err)
        
                            if scp_process.returncode!=0:
                                if finished:
                                    status='ERR'
                                    self.log.error('download failed: %s (%s)', destination, scp_out+scp_err)
                                    self.log.debug('epoch=%d file=%s', esx_epoch, line)
                                else:
                                    self.log.info('download failed, retry later %s (%s)', destination, scp_out+scp_err)
                    
                    if not finished:
                        # reload the list to work on fresh data
                        wait=1 
                        break
                    


            end=time.time()                
            if status!='OK':
                self.upload_status_code='ERR'
            self.upload_status+='upload_status-%s=%s\r\n' % (current_vm, status)
            self.upload_status+='upload_end-%s=%s\r\n' % (current_vm, time.ctime(end))
            self.upload_status+='upload_end_epoch-%s=%d\r\n' % (current_vm, end)
            
            if target and type=='move':
                # remove source
                exit_code, output=RunCommand(self.t, 'rm -rf "%s"' % (find_src, ), log=None)
                if exit_code:
                    self.log.error('failed to remove %s: %r', find_src, output)
                    self.upload_status_code='ERR'
                else:
                    self.log.info('remove %s', find_src)

                            
        self.upload_status='upload_status=%s\r\n%s' % (self.upload_status_code, self.upload_status)
        return True

    # -----------------------------------------------------------------------
    def run(self, command, job, manager):

        self.log=log=manager.log
        self.manager=manager
        
        now=manager.now
        log.info('ghettoVCB.sh version is: %s  nfs_wait=%r', self.ghettovcb_version, self.ghettovcb_wait_nfs)
        start=int(time.time())

        self.t=t=paramiko.Transport((self.host, self.port))
        # t.get_security_options()._set_ciphers(['blowfish-cbc', ]) # don't speed up anything and don't work with 3.5 
        t.connect(username=self.login, password=self.password)
        # t.set_keepalive(300) # make the backup fail
        chan=t.open_session()
        chan.set_combine_stderr(True)
        log.info('start: %s', self.cmd_line)
        chan.exec_command(self.cmd_line)

        attachments=[]
        self.finished_vm=set()
        output=''
        state='waiting_for_start'
        current_vm=None
        for line in socket_readlines(chan):
            # I'm reading the socket in this loop,
            # I CANNOT DO SOMETHING TIME CONSUMING or do it in background 
            output+=line+'\n'
            log.debug(line)
            
            match=self.config_vm_backup_volume_re.match(line)
            if match:
                vm_backup_volume=match.group('vm_backup_volume')
                continue

            match=self.config_disk_backup_format_re.match(line)
            if match:
                disk_backup_format=match.group('disk_backup_format')
                continue

            match=self.config_vm_backup_dir_naming_convention_re.match(line)
            if match:
                vm_backup_dir_naming_convention=match.group('vm_backup_dir_naming_convention')
                continue
            
            match=self.initiate_backup_re.match(line)
            if match:
                # start a new VM
                state='started'
                current_vm=match.group('name')

                # read vm config file if it exist
                filename=os.path.join(self.local, current_vm)
                if os.path.isfile(filename):
                    vm_var=self.read_vm_conf(filename, self.default_var)
                else:
                    vm_var=self.default_var
                    
                # use vm_backup_volume instead of vm_var['VM_BACKUP_VOLUME']
                # because the first one as been recalculated by ghettoVCB to take care of 
                # ENABLE_NON_PERSISTENT_NFS
                target=posixpath.join(vm_backup_volume, current_vm)
                compress=vm_var['ENABLE_COMPRESSION']=='1'
                persistent_nfs=vm_var['ENABLE_NON_PERSISTENT_NFS']!='1'
                root_backup_dir=posixpath.join(vm_backup_volume, current_vm, current_vm+'-'+vm_backup_dir_naming_convention)
                # I cannot start upload_vm now, because of a race condition, I need to wait for
                # the start of first disk to besure direcrories exists
                wait_first_disk=True
                continue

            match=self.find_vmdk_re.match(line)
            if match:
                # start a new disk
                if wait_first_disk and self.destination and not compress and persistent_nfs:
                    # monitor directory and an upload "finished file" file 
                    self.upload_vm(current_vm, root_backup_dir, now, False)
                wait_first_disk=False
                continue

            match=self.backup_duration_re.match(line)
            if match:
                # the backup just finished
                self.finished_vm.add(current_vm)
                if self.destination and compress and persistent_nfs:
                    # upload compressed file 
                    self.upload_vm(current_vm, posixpath.dirname(root_backup_dir), now, posixpath.basename(root_backup_dir)+'.gz')
            
                # search the target and make a directory listing 
                dir_lst, up_lst, scp_process=None, None, None
                if not persistent_nfs and not self.ghettovcb_wait_nfs:
                    log.warning('cannot retrieve data from non persistent NFS share on old version of ghettoVCB.sh')
                else:
                    if compress:
                        _exit_code, dir_lst=RunCommand(t, 'ls -l "%s"' % target, log)
                        dir_lst='== directory '+target+'\n'+dir_lst
                    else:
                        _exit_code, up_lst=RunCommand(t, 'ls "%s"' % target, log)
                        dir_lst=None
                        # search the target directory, take the last one
                        target_name=''
                        for line in up_lst.split('\n'):
                            if line.endswith('-symlink'):
                                continue
                            if line.startswith(current_vm):
                                if line>target_name:
                                   target_name=line
                        up_lst='== directory '+target+'\n'+up_lst
                        if target_name: 
                            target=posixpath.join(target, target_name)
                            _exit_code, dir_lst=RunCommand(t, 'ls -l "%s"' % target, log)
                            dir_lst='== directory '+target+'\n'+dir_lst
    
                    if dir_lst:
                        attachments.append(('dir-%s.txt' % (current_vm,), None, dir_lst, 'text', 'plain', 'utf-8'))
            
                    if up_lst:
                        attachments.append(('dirup-%s.txt' % (current_vm,), None, up_lst, 'text', 'plain', 'utf-8'))
        
        ghettovcb_exit_code=chan.recv_exit_status()
        chan.close()
        # log.debug('ghettovcb return unreliable ssh exit code: %d', ghettovcb_exit_code)
        end_ghettovcb=int(time.time())
        
        # ESX(i) (dropbear) returns an unreliable exit_code   
        pos1=output.rfind('exit_code=')
        pos2=output.find('\n', pos1)
        ghettovcb_exit_code=int(output[pos1+10:pos2])
        output=output[:pos1]

        attachments.insert(0, ('output.txt', None, output, 'text', 'plain', 'utf-8') )
        
        # make the disk space usage
        _exit_code, df=RunCommand(t, 'df', log)
        
        # make volume listing 
        _exit_code, ls_volumes=RunCommand(t, 'ls -l /vmfs/volumes', log)
        ls_volumes='== directory /vmfs/volumes\n'+ls_volumes

        # Nothing more to upload    
        if self.upload_thread:
            self.upload_vm(None, None ,None, None)
            # wait for end of upload/download
            log.info('waiting for end of upload...')
            self.upload_thread.join()
            if self.download_script:
                attachments.insert(0, ('download.txt', None, self.download_script, 'text', 'plain', 'utf-8') )

        end=int(time.time())

        t.close()

        if ghettovcb_exit_code==0:
            ghettoVCB_status='OK'
            backup_status='OK'
        else:
            ghettoVCB_status='ERR'
            backup_status='ERR'

        if self.upload_status_code and self.upload_status_code!='OK':
            backup_status='ERR'

        status=u''
        status+='name=%s\r\n' % job['name']
        status+='program=%s\r\n' % self.name
        status+='version=%s\r\n' % self.__version__
        status+='status=%s\r\n' % backup_status
        status+='hostname=%s\r\n' % manager.hostname
        
        status+='exit_code=%d\r\n' % ghettovcb_exit_code
        status+='start=%s\r\n' % time.ctime(start)
        status+='end_ghettovcb=%s\r\n' % time.ctime(end_ghettovcb)
        status+='end=%s\r\n' % time.ctime(end)
        status+='start_epoch=%d\r\n' % start
        status+='end_epoch=%d\r\n' % end
        status+='end_ghettovcb_epoch=%d\r\n' % end_ghettovcb

        status+='type=%s\r\n' % self.type
        status+='ghettovcb_status=%s\r\n' % ghettoVCB_status
        status+=self.upload_status

        if df:
            attachments.insert(0, ('df.txt', None, df, 'text', 'plain', 'utf-8'),)

        if ls_volumes:
            attachments.insert(0, ('ls_volumes.txt', None, ls_volumes, 'text', 'plain', 'utf-8'),)
        
        return backup_status, status, attachments



#
# ntbackup.py
#
# ntbackup manager for mkbackup
# 
# (c) alain.spineux@gmail.com

import os, sys, subprocess, time, platform
from datetime import datetime, timedelta
import win32com.shell
from win32com.shell import shell, shellcon

from archiver import *
from windows import WindowsErrorDecode, windows_list_dir

import cron


# ===========================================================================
class NTBackup(Archiver):
    # http://support.microsoft.com/kb/814583
    # http://support.microsoft.com/default.aspx?scid=kb;en-us;233427
    #
    # Normal - selected files, marking their archive attributes
    # Copy - selected files without marking their archive attributes. This is good for making tape copies that do not interfere with archive backups, since it does not set the archive attribute.
    # Incremental - selected files, marking their archive attributes, but only backs up the ones that have changed since the last backup.
    # Differential - selected files, NOT marking their archive attributes, but only backs up the ones that have changed since the last backup.
    # Daily - only backs up files that have changed that day, does not mark their archive attributes.
    #
    # ntbackup stuff
    #   http://www.fishbrains.com/2007/11/12/utilizing-the-built-in-windows-backup-ntbackupexe-for-windows/
    #   http://episteme.arstechnica.com/eve/forums/a/tpc/f/12009443/m/165002540831
    
    name='ntbackup'
    exe='ntbackup.exe'
    ev_logtype='Application'
    ev_source='NTBackup'
    
    boolean={ True: 'yes', False:'no'}
    onoff={ True: 'on', False:'off'}
    
    types=dict(normal='normal', full='normal', incremental='incremental', inc='incremental', differential='differential', diff='differential', copy='copy', daily='daily')

    # -----------------------------------------------------------------------
    def __init__(self):
        Archiver.__init__(self)
        
        self.selection=None
        self.verify=None
        self.restricted=None
        self.logdir=None
        self.snap=None
        self.options=[]
        
        self.type=None
        self.target=None
        self.target_dir=None
        
        systemroot=os.environ.get('SystemRoot', os.environ.get('windir', 'C:\\windows'))
        self.program_exe=os.path.join(systemroot, 'system32', 'ntbackup.exe')

        self.is64bits=os.path.isdir(os.path.join(systemroot, 'SysWOW64'))            

        # check if we are on a x64 running a 32bit MKBackup
        
    # -----------------------------------------------------------------------
    def load(self, job, manager):
        
        log=manager.log
        now=manager.now

        errors, warnings, extra={}, {}, ''

        #
        # check job config
        #
        
        self.selection=job.get('selection', None)
        if not self.selection:
            errors['selection']='option mandatory'
        else:
            if self.selection.startswith('"') and self.selection.endswith('"'):
                self.selection=self.selection[1:-1]
            if self.selection.startswith('@'):
                if not os.path.isfile(self.selection[1:]):
                    errors['selection']='file not found: %s' % (self.selection[1:],)
            elif os.path.isfile(self.selection) and self.selection[-4:].lower()=='.bks':
                errors['selection']='looks like a ".bks" file and need to be prefixed by a @'
            elif not os.path.isdir(self.selection) and not os.path.isfile(self.selection):
                errors['selection']='file or directory not found'

        self.verify=boolean.get(job.get('verify', 'no').lower(), None)
        if self.verify==None:
            errors['verify']='boolean must have value in (on, yes, true, 1, off, no, false and 0)'

        if job.get('snap'):
            self.snap=boolean.get(job.get('snap').lower(), None)
        if self.verify==None:
            errors['snap']='boolean must have value in (on, yes, true, 1, off, no, false and 0)'

        self.restricted=boolean.get(job.get('restricted', 'no').lower(), None)
        if self.restricted==None:
            errors['restricted']='boolean must have value in (on, yes, true, 1, off, no, false and 0)'

        self.logdir=job.get('logdir', None)
        if not self.logdir:
            self.logdir=os.path.join(shell.SHGetPathFromIDList(shell.SHGetSpecialFolderLocation (0, shellcon.CSIDL_LOCAL_APPDATA)), 'Microsoft\Windows NT\NTBackup\data')
            # dont check if logdir exist here because ntbackup.exe will create it
        else:
            if self.logdir.startswith('"') and self.logdir.endswith('"'):
                self.logdir=self.logdir[1:-1]

        if not os.path.isdir(self.logdir):
            # I want to be sure the user know what he is doing
            warnings['logdir']='directory not found ! Notice, NTBackup create its "logdir" directory at first run'
    
        self.program_exe=job.get('ntbackup', self.program_exe)
        if self.program_exe.startswith('"') and self.program_exe.endswith('"'):
            self.program_exe=self.program_exe[1:-1]
        
        if not os.path.exists(self.program_exe):
            if self.is64bits:
                errors['ntbackup']='file "%s" not found ! On 64bits system you must use the 64bits trick, see the doc for more' % (self.program_exe, )
            else:
                errors['ntbackup']='file "%s" not found ! ' % (self.program_exe, )

        self.destination=job.get('destination', None)
        if not self.destination:
            errors['destination']='option mandatory'
        else:
            self.destination=self.destination.replace('\n','')
            try:
                self.destination=Destinations(self.destination, self)
            except (DestinationSyntaxError, cron.CronException), e:
                errors['destination']='syntax error: %s' % (str(e), )
            else:
                self.type, self.target=self.destination.match(now, self.night_shift)
                if self.type!=None and self.type!='none':
                    self.target_dir=os.path.dirname(self.target)
                    if not os.path.isdir(self.target_dir):
                        try:
                            os.makedirs(self.target_dir)
                        except WindowsError, e:
                            errors['destination']='directory not found: %s (%s)' % (self.target_dir, WindowsErrorDecode(e))

        options=job.get('options', '')
        try:
            options=quoted_string_list(options)
        except ValueError:
            errors['options']='not a valid quoted string list'
        else:
            self.options=options

        
        extra+='ntbackup=%s\n' % self.program_exe

        if not errors:
    
            if self.type!=None and self.type!='none':
                
                # ntbackup backup "@m:\asx\src\magik\job1.bks" /a /d "Set created 14/11/2009 at 2:29" 
                # /v:no /r:no /rs:no /hc:off /m normal /j "backup" /l:s /f "s:\Backup.bkf"
                self.cmd_args=[ self.program_exe, 'backup', self.selection, '/J', job['name'], '/M', self.type, '/F', self.target, '/rs:no', ]
                if job.get('description', None): args.extend(['/d', job.get('description')])
                self.cmd_args.append('/v:'+self.boolean[self.verify])
                self.cmd_args.append('/r:'+self.boolean[self.restricted])
                if self.snap:
                    self.cmd_args.append('/SNAP:'+self.onoff[self.snap])
                
                self.cmd_args.append('/l:s') # logging [s]ummary
                self.cmd_args.extend(self.options)
                # '/hc:off'
            
                extra+='cmdline=%s\n' % '  '.join(self.cmd_args)
    
        return errors, warnings, extra
        
    # -----------------------------------------------------------------------
    def run(self, command, job, manager):

        log=manager.log
        now=manager.now

        import windows

        try:
            start=int(time.time())
            process=subprocess.Popen(param_encode(self.cmd_args, manager.default_encoding), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            std_out, std_err=process.communicate()
            end=int(time.time())
        except WindowsError:
            log.error('command: %r', self.cmd_args)
            if self.is64bits: 
                log.error('On 64bits system you must use the 64bits trick, check the doc for more...')
            raise

        # choose the oldest log file
        logfile, last=None, 0
        for filename in os.listdir(self.logdir):
            full_filename=os.path.join(self.logdir, filename)
            current=os.stat(full_filename).st_mtime
            if last<current:
                logfile, last=full_filename, current
                
        # retrieve records in the event viewer
        ev_status, ev_out=windows.ReadEvLog(NTBackup.ev_logtype, NTBackup.ev_source, log, start)
        if not ev_out:
            log.warning('event log is empty')

        if logfile:
            if ev_status=='OK':
                flog=log.debug
            else:
                flog=log.info
            for line in open(logfile, 'r').read().decode('utf-16').split('\n'):
                flog('> %s', line)
        
        status=u''
        status+='name=%s\r\n' % job['name']
        status+='program=%s\r\n' % self.name
        status+='version=%s\r\n' % self.__version__
        status+='hostname=%s\r\n' % platform.node()
        
        status+='status=%s\r\n' % ev_status

        status+='exit_code=%r\r\n' % process.returncode
        status+='start=%s\r\n' % time.ctime(start)
        status+='end=%s\r\n' % time.ctime(end)
        status+='logfile=%s\r\n' % logfile
        
        status+='start_epoch=%d\r\n' % start
        status+='end_epoch=%d\r\n' % end
        status+='selection=%s\r\n' % self.selection
        status+='type=%s\r\n' % self.type
        status+='target=%s\r\n' % self.target
        
        # retrieve info about the target file
        try:
            stat=os.stat(self.target)
        except Exception, e:
            log.error('checking target file: %s', WindowsErrorDecode(e))
            status+='target_stat_error=%s\r\n' % WindowsErrorDecode(e)
        else:
            status+='target_size=%d\r\n' % stat.st_size
            status+='target_mtime_epoch=%d\r\n' % stat.st_mtime
            status+='target_mtime=%s\r\n' % time.ctime(stat.st_mtime)
        
        # retrieve info about the target directory
        dir_out=None
        try:
            total_free_bytes=free_space(self.target_dir, log)
            dir_out, _total_size=windows_list_dir(self.target_dir, log)        
        except Exception, e:
            log.error('checking target directory: %s', WindowsErrorDecode(e))
            status+='target_directory_error=%s\r\n' % WindowsErrorDecode(e)
        else:
            status+='target_free_space=%d\r\n' % total_free_bytes

        status+='cmdline=%s\r\n' % ' '.join(self.cmd_args)

        attachments=[ # the type, subtype and coding is not used
                       ('log.txt', logfile, None, 'text', 'plain', 'utf-16'),
                       ('evlog.txt', None, ev_out.encode('utf-8'), 'text', 'plain', 'utf-8'),
                    ]

        if dir_out:
            attachments.append(('dir.txt', None, dir_out.encode('utf-8'), 'text', 'plain', 'utf-8'))
        if self.selection.startswith('@'):
            attachments.append(('selection.bks', self.selection[1:], None, 'text', 'plain', 'utf-16'))
        if std_out:
            attachments.append(('stdout.txt', None, std_out, 'text', 'plain', manager.console_encoding))
        if std_err:
            attachments.append(('stderr.txt', None, std_err, 'text', 'plain', manager.console_encoding))

        return ev_status, status, attachments



#
# (C) Copyright 2003-2007 Hewlett-Packard Development Company, L.P.
# (C) Copyright 2006-2007 Novell, Inc. 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Tim Potter <tpot@hp.com>
#         Martin Pool <mbp@hp.com>
#         Bart Whiteley <bwhiteley@suse.de>

import string, re
import cim_xml, cim_types
from types import StringTypes
from cim_types import atomic_to_cim_xml, CIMDateTime
from cim_xml import *
from datetime import datetime, timedelta

"""
Representations of CIM Objects.

In general we try to map CIM objects directly into Python primitives,
except when that is not possible or would be ambiguous.  For example,
CIM Class names are simply Python strings, but a ClassPath is
represented as a special Python object.

These objects can also be mapped back into XML, by the toxml() method
which returns a string.
"""

class NocaseDict(object):
    """Yet another implementation of a case-insensitive dictionary."""

    def __init__(self, *args, **kwargs):

        self.data = {}

        # Initialise from sequence object

        if len(args) == 1 and type(args[0]) == list:
            for item in args[0]:
                self[item[0]] = item[1]

        # Initialise from mapping object

        if len(args) == 1 and type(args[0]) == dict:
            self.update(args[0])

        # Initialise from NocaseDict

        if len(args) == 1 and isinstance(args[0], NocaseDict):
            self.data = args[0].data.copy()

        # Initialise from keyword args

        self.update(kwargs)

    # Basic accessor and settor methods

    def __getitem__(self, key):
        k = key
        if isinstance(key, (str, unicode)):
            k = key.lower()
        return self.data[k][1]

    def __setitem__(self, key, value):
        if not isinstance(key, (str, unicode)):
            raise KeyError, 'Key must be string type'
        k = key.lower()
        self.data[k] = (key, value)

    def __delitem__(self, key):
        k = key
        if isinstance(key, (str, unicode)):
            k = key.lower()
        del self.data[k]

    def __len__(self):
        return len(self.data)

    def has_key(self, key):
        k = key
        if isinstance(key, (str, unicode)):
            k = key.lower()
        return self.data.has_key(k)

    def __contains__(self, key):
        k = key
        if isinstance(key, (str, unicode)):
            k = key.lower()
        return k in self.data

    def get(self, key, default = None):
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key, default):
        if not self.has_key(key):
            self[key] = default
        return self[key]

    # Other accessor expressed in terms of iterators

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    # Iterators

    def iterkeys(self):
        for item in self.data.iteritems():
            yield item[1][0]

    def itervalues(self):
        for item in self.data.iteritems():
            yield item[1][1]

    def iteritems(self):
        for item in self.data.iteritems():
            yield item[1]

    # Other stuff

    def __repr__(self):
        items = ', '.join([('%r: %r' % (key, value))
                           for key, value in self.items()])
        return 'NocaseDict({%s})' % items

    def update(self, *args, **kwargs):
        for mapping in args:
            if hasattr(mapping, 'items'):
                for k, v in mapping.items():
                    self[k] = v
            else:
                for (k, v) in mapping:
                    self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def clear(self):
        self.data.clear()

    def popitem(self):
        pass

    def copy(self):
        result = NocaseDict()
        result.data = self.data.copy()
        return result

    def __eq__(self, other):
        for key, value in self.iteritems():
            if not (key in other) or not (other[key] == value):
                return 0
        return len(self) == len(other)

    def __cmp__(self, other):
        for key, value in self.iteritems():
            if not (key in other):
                return -1
            rv = cmp(value, other[key])
            if rv != 0:
                return rv
        return len(self) - len(other)

def cmpname(name1, name2):
    """Compare to CIM names.  The comparison is done
    case-insensitvely, and one or both of the names may be None."""

    if name1 is None and name2 is None:
        return 0

    if name1 is None:
        return -1

    if name2 is None:
        return 1

    lower_name1 = name1.lower()
    lower_name2 = name2.lower()

    return cmp(lower_name1, lower_name2)

class CIMClassName(object):

    def __init__(self, classname, host = None, namespace = None):

        if not isinstance(classname, StringTypes):
            raise TypeError('classname argument must be a string')

        # TODO: There are some odd restrictions on what a CIM
        # classname can look like (i.e must start with a
        # non-underscore and only one underscore per classname).

        self.classname = classname
        self.host = host
        self.namespace = namespace
        
    def copy(self):
        return CIMClassName(self.classname, host = self.host,
                            namespace = self.namespace)

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, CIMClassName):
            return 1

        return (cmpname(self.classname, other.classname) or
                cmpname(self.host, other.host) or
                cmpname(self.namespace, other.namespace))

    def __str__(self):
        
        s = ''

        if self.host is not None:
            s += '//%s/' % self.host

        if self.namespace is not None:
            s += '%s:' % self.namespace

        s += self.classname

        return s

    def __repr__(self):

        r = '%s(classname=%s' % (self.__class__.__name__, `self.classname`)

        if self.host is not None:
            r += ', host=%s' % `self.host`

        if self.namespace is not None:
            r += ', namespace=%s' % `self.namespace`

        r += ')'

        return r

    def tocimxml(self):

        classname = cim_xml.CLASSNAME(self.classname)

        if self.namespace is not None:

            localnsp = cim_xml.LOCALNAMESPACEPATH(
                [cim_xml.NAMESPACE(ns)
                 for ns in string.split(self.namespace, '/')])

            if self.host is not None:

                # Classname + namespace + host = CLASSPATH

                return cim_xml.CLASSPATH(
                    cim_xml.NAMESPACEPATH(cim_xml.HOST(self.host), localnsp),
                    classname)

            # Classname + namespace = LOCALCLASSPATH
                                          
            return cim_xml.LOCALCLASSPATH(localnsp, classname)

        # Just classname = CLASSNAME
        
        return cim_xml.CLASSNAME(self.classname)

class CIMProperty(object):
    """A property of a CIMInstance.

    Property objects represent both properties on particular instances,
    and the property defined in a class.  In the first case, the property
    will have a Value and in the second it will not.

    The property may hold an array value, in which case it is encoded
    in XML to PROPERTY.ARRAY containing VALUE.ARRAY."""
    
    def __init__(self, name, value, type = None, 
                 class_origin = None, array_size = None, propagated = None,
                 is_array = False, reference_class = None, qualifiers = {},
                 embedded_object = None):

        # Initialise members

        self.name = name
        self.value = value
        self.type = type
        self.class_origin = class_origin
        self.array_size = array_size
        self.propagated = propagated
        self.qualifiers = NocaseDict(qualifiers)
        self.is_array = is_array
        self.reference_class = reference_class
        self.embedded_object = embedded_object
        
        if isinstance(value, (datetime, timedelta)):
            value = CIMDateTime(value)

        import __builtin__
        if __builtin__.type(value) == list:
            self.is_array = True
        # Determine type of value if not specified

        if type is None:

            # Can't work out what is going on if type and value are
            # both not set.

            if value is None:
                raise TypeError('Null property "%s" must have a type' % name)
        
            if self.is_array:

                # Determine type for list value

                if len(value) == 0:
                    raise TypeError(
                        'Empty property array "%s" must have a type' % name)
                    
                elif isinstance(value[0], CIMInstance):
                    self.type = 'string'
                    self.embedded_object = 'instance'
                elif isinstance(value[0], CIMClass):
                    self.type = 'string'
                    self.embedded_object = 'object'
                else:
                    self.type = cim_types.cimtype(value[0])

            elif isinstance(value, CIMInstanceName):

                self.type = 'reference'

            elif isinstance(value, CIMInstance):
                self.type = 'string'
                self.embedded_object = 'instance'

            elif isinstance(value, CIMClass):
                self.type = 'string'
                self.embedded_object = 'object'
                
            else:

                # Determine type for regular value
                
                self.type = cim_types.cimtype(value)

    def copy(self):

        return CIMProperty(self.name,
                           self.value,
                           type = self.type,
                           class_origin = self.class_origin,
                           array_size = self.array_size,
                           propagated = self.propagated,
                           is_array = self.is_array,
                           reference_class = self.reference_class,
                           qualifiers = self.qualifiers.copy())

    def __repr__(self):

        return '%s(name=%s, type=%s, value=%s, is_array=%s)' % \
               (self.__class__.__name__, `self.name`, `self.type`,
                `self.value`, `self.is_array`)

    def tocimxml(self):

        if self.is_array:

            value = self.value
            if value is not None:
                if value:
                    if self.embedded_object is not None:
                        value = [v.tocimxml().toxml() for v in value]
                value = VALUE_ARRAY([VALUE(atomic_to_cim_xml(v)) for v in value])

            return PROPERTY_ARRAY(
                self.name,
                self.type,
                value,
                self.array_size,
                self.class_origin,
                self.propagated,
                qualifiers = [q.tocimxml() for q in self.qualifiers.values()],
                embedded_object = self.embedded_object)

        elif self.type == 'reference':

            value_reference = None
            if self.value is not None:
                value_reference = VALUE_REFERENCE(self.value.tocimxml())

            return PROPERTY_REFERENCE(
                self.name,
                value_reference,
                reference_class = self.reference_class,
                class_origin = self.class_origin,
                propagated = self.propagated,
                qualifiers = [q.tocimxml() for q in self.qualifiers.values()])

        else:
            value = self.value
            if value is not None:
                if self.embedded_object is not None:
                    value = value.tocimxml().toxml()
                else:
                    value = atomic_to_cim_xml(value)
                value = VALUE(value)

            return PROPERTY(
                self.name,
                self.type,
                value,
                class_origin = self.class_origin,
                propagated = self.propagated,
                qualifiers = [q.tocimxml() for q in self.qualifiers.values()],
                embedded_object = self.embedded_object)

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, self.__class__):
            return 1

        return (cmpname(self.name, other.name)
                or cmp(self.value, other.value)
                or cmp(self.type, other.type)
                or cmp(self.class_origin, other.class_origin)
                or cmp(self.array_size, other.array_size)
                or cmp(self.propagated, other.propagated)
                or cmp(self.qualifiers, other.qualifiers)
                or cmp(self.is_array, other.is_array)
                or cmpname(self.reference_class, other.reference_class))

class CIMInstanceName(object):
    """Name (keys) identifying an instance.

    This may be treated as a dictionary to retrieve the keys."""

    def __init__(self, classname, keybindings = {}, host = None,
                 namespace = None):

        self.classname = classname
        self.keybindings = NocaseDict(keybindings)
        self.host = host
        self.namespace = namespace

    def copy(self):

        result = CIMInstanceName(self.classname)
        result.keybindings = self.keybindings.copy()
        result.host = self.host
        result.namespace = self.namespace
        
        return result

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, CIMInstanceName):
            return 1

        return (cmpname(self.classname, other.classname) or
                cmp(self.keybindings, other.keybindings) or
                cmpname(self.host, other.host) or
                cmpname(self.namespace, other.namespace))

    def __str__(self):

        s = ''

        if self.host is not None:
            s += '//%s/' % self.host

        if self.namespace is not None:
            s += '%s:' % self.namespace

        s += '%s.' % self.classname

        for key, value in self.keybindings.items():

            s +='%s=' % key

            if type(value) == int or type(value) == long:
                s += str(value)
            else:
                s += '"%s"' % value

            s += ','
            
        return s[:-1]

    def __repr__(self):

        r = '%s(classname=%s, keybindings=%s' % \
            (self.__class__.__name__,
             `self.classname`,
             `self.keybindings`)

        if self.host is not None:
            r += ', host=%s' % `self.host`

        if self.namespace is not None:
            r += ', namespace=%s' % `self.namespace`

        r += ')'

        return r
        
    # A whole bunch of dictionary methods that map to the equivalent
    # operation on self.keybindings.

    def __getitem__(self, key): return self.keybindings[key]
    def __contains__(self, key): return key in self.keybindings
    def __delitem__(self, key): del self.keybindings[key]
    def __setitem__(self, key, value): self.keybindings[key] = value
    def __len__(self): return len(self.keybindings)
    def has_key(self, key): return self.keybindings.has_key(key)
    def keys(self): return self.keybindings.keys()
    def values(self): return self.keybindings.values()
    def items(self): return self.keybindings.items()
    def iterkeys(self): return self.keybindings.iterkeys()
    def itervalues(self): return self.keybindings.itervalues()
    def iteritems(self): return self.keybindings.iteritems()
    def update(self, *args, **kwargs): self.keybindings.update(*args, **kwargs)

    def tocimxml(self):

        # Generate an XML representation of the instance classname and
        # keybindings.

        if type(self.keybindings) == str:

            # Class with single key string property
        
            instancename_xml = cim_xml.INSTANCENAME(
                self.classname,
                cim_xml.KEYVALUE(self.keybindings, 'string'))

        elif isinstance(self.keybindings, (long, float, int)): 

            # Class with single key numeric property
        
            instancename_xml = cim_xml.INSTANCENAME(
                self.classname,
                cim_xml.KEYVALUE(str(self.keybindings), 'numeric'))

        elif isinstance(self.keybindings, (dict, NocaseDict)):

            # Dictionary of keybindings
            # NOCASE_TODO should remove dict below. 

            kbs = []

            for kb in self.keybindings.items():

                # Keybindings can be integers, booleans, strings or
                # value references.                

                if hasattr(kb[1], 'tocimxml'):
                    kbs.append(cim_xml.KEYBINDING(
                        kb[0],
                        cim_xml.VALUE_REFERENCE(kb[1].tocimxml())))
                    continue
                               
                if type(kb[1]) == bool:
                    _type = 'boolean'
                    if kb[1]:
                        value = 'TRUE'
                    else:
                        value = 'FALSE'
                elif isinstance(kb[1], (long, float, int)): 
                    # pywbem.cim_type.{Sint32, Real64, ... } derive from 
                    # long or float
                    _type = 'numeric'
                    value = str(kb[1])
                elif type(kb[1]) == str or type(kb[1]) == unicode:
                    _type = 'string'
                    value = kb[1]
                else:
                    raise TypeError(
                        'Invalid keybinding type for keybinding ' '%s: %s' % (kb[0],`type(kb[1])`))

                kbs.append(cim_xml.KEYBINDING(
                    kb[0],
                    cim_xml.KEYVALUE(value, _type)))

            instancename_xml = cim_xml.INSTANCENAME(self.classname, kbs)

        else:

            # Value reference

            instancename_xml = cim_xml.INSTANCENAME(
                self.classname,
                cim_xml.VALUE_REFERENCE(self.keybindings.tocimxml()))

        # Instance name plus namespace = LOCALINSTANCEPATH

        if self.host is None and self.namespace is not None:
            return cim_xml.LOCALINSTANCEPATH(
                cim_xml.LOCALNAMESPACEPATH(
                    [cim_xml.NAMESPACE(ns)
                     for ns in string.split(self.namespace, '/')]),
                instancename_xml)

        # Instance name plus host and namespace = INSTANCEPATH

        if self.host is not None and self.namespace is not None:
            return cim_xml.INSTANCEPATH(
                cim_xml.NAMESPACEPATH(
                    cim_xml.HOST(self.host),
                    cim_xml.LOCALNAMESPACEPATH([
                        cim_xml.NAMESPACE(ns)
                        for ns in string.split(self.namespace, '/')])),
                instancename_xml)

        # Just a regular INSTANCENAME

        return instancename_xml

class CIMInstance(object):
    """Instance of a CIM Object.

    Has a classname (string), and named arrays of properties and qualifiers.

    The properties is indexed by name and points to CIMProperty
    instances."""

    def __init__(self, classname, properties = {}, qualifiers = {},
                 path = None, property_list = None):
        """Create CIMInstance.

        bindings is a concise way to initialize property values;
        it is a dictionary from property name to value.  This is
        merely a convenience and gets the same result as the
        properties parameter.

        properties is a list of full CIMProperty objects. """
        
        self.classname = classname
        self.qualifiers = NocaseDict(qualifiers)
        self.path = path
        if property_list is not None:
            self.property_list = [x.lower() for x in property_list]
        else:
            self.property_list = None

        # Assign initialised property values and run through
        # __setitem__ to enforce CIM types for each property.

        self.properties = NocaseDict()
        [self.__setitem__(k, v) for k, v in properties.items()]

    def update(self, *args, **kwargs):
        """D.update(E, **F) -> None.  
        
        Update D from E and F: for k in E: D[k] = E[k]
        (if E has keys else: for (k, v) in E: D[k] = v) 
        then: for k in F: D[k] = F[k] """

        for mapping in args:
            if hasattr(mapping, 'items'):
                for k, v in mapping.items():
                    self[k] = v
            else:
                for (k, v) in mapping:
                    self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def update_existing(self, *args, **kwargs):
        """Update property values iff the property previously exists.
        
        Update D from E and F: for k in E: D[k] = E[k]
        (if E has keys else: for (k, v) in E: D[k] = v) 
        then: for k in F: D[k] = F[k] 
        
        Like update, but properties that are not already present in the 
        instance are skipped. """

        for mapping in args:
            if hasattr(mapping, 'items'):
                for k, v in mapping.items():
                    try:
                        prop = self.properties[k]
                    except KeyError:
                        continue
                    prop.value = tocimobj(prop.type, v)
            else:
                for (k, v) in mapping:
                    try:
                        prop = self.properties[k]
                    except KeyError:
                        continue
                    prop.value = tocimobj(prop.type, v)
        for k, v in kwargs.items():
            try:
                prop = self.properties[k]
            except KeyError:
                continue
            prop.value = tocimobj(prop.type, v)

    def copy(self):

        result = CIMInstance(self.classname)
        result.properties = self.properties.copy()
        result.qualifiers = self.qualifiers.copy()
        result.path = (self.path is not None and [self.path.copy()] or [None])[0]

        return result

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, CIMInstance):
            return 1

        return (cmpname(self.classname, other.classname) or
                cmp(self.path, other.path) or
                cmp(self.properties, other.properties) or
                cmp(self.qualifiers, other.qualifiers))

    def __repr__(self):
        # Don't show all the properties and qualifiers because they're
        # just too big
        return '%s(classname=%s, ...)' % (self.__class__.__name__,
                                          `self.classname`)

    # A whole bunch of dictionary methods that map to the equivalent
    # operation on self.properties.

    def __contains__(self, key):
        return key in self.properties

    def __getitem__(self, key):
        return self.properties[key].value
    
    def __delitem__(self, key):
        del self.properties[key]
        
    def __len__(self):
        return len(self.properties)
    
    def has_key(self, key):
        return self.properties.has_key(key)
    
    def keys(self):
        return self.properties.keys()
    
    def values(self):
        return [v.value for v in self.properties.values()]
    
    def items(self):
        return [(k, v.value) for k, v in self.properties.items()]
    
    def iterkeys(self):
        return self.properties.iterkeys()
    
    def itervalues(self):
        for k, v in self.properties.iteritems():
            yield v.value
    
    def iteritems(self):
        for k, v in self.properties.iteritems():
            yield (k, v.value)
    
    def __setitem__(self, key, value):

        # Don't let anyone set integer or float values.  You must use
        # a subclass from the cim_type module.

        if type(value) == int or type(value) == float or type(value) == long:
            raise TypeError('Must use a CIM type assigning numeric values.')

        if self.property_list is not None and key.lower() not in \
                self.property_list:
            if self.path is not None and key not in self.path.keybindings:
                return
        # Convert value to appropriate PyWBEM type

        if isinstance(value, CIMProperty):
            v = value
        else:
            v = CIMProperty(key, value)

        self.properties[key] = v
        if self.path is not None and key in self.path.keybindings:
            self.path[key] = v.value
        
    def tocimxml(self):

        props = []

        for key, value in self.properties.items():
            
            # Value has already been converted into a CIM object
            # property type (e.g for creating null property values).

            if isinstance(value, CIMProperty):
                props.append(value)
                continue

            props.append(CIMProperty(key, value))

        instance_xml = cim_xml.INSTANCE(
            self.classname,
            properties = [p.tocimxml() for p in props],
            qualifiers = [q.tocimxml() for q in self.qualifiers.values()])

        if self.path is None:
            return instance_xml

        return cim_xml.VALUE_NAMEDINSTANCE(self.path.tocimxml(),
                                           instance_xml)

    def tomof(self):

        def _prop2mof(_type, value):
            if value is None:
                val = 'NULL'
            elif isinstance(value, list):
                val = '{'
                for i,x in enumerate(value):
                    if i > 0:
                        val += ', '
                    val += _prop2mof(_type, x)
                val += '}'
            elif _type == 'string':
                val = '"' + value + '"'
            else:
                val = str(value)
            return val

        s = 'instance of %s {\n' % self.classname
        for p in self.properties.values():
            s+= '\t%s = %s;\n' % (p.name, _prop2mof(p.type, p.value))
        s+= '};\n'

        return s



class CIMClass(object):

    def __init__(self, classname, properties = {}, methods = {},
                 superclass = None, qualifiers = {}):

        self.classname = classname
        self.properties = NocaseDict(properties)
        self.qualifiers = NocaseDict(qualifiers)
        self.methods = NocaseDict(methods)
        self.superclass = superclass

    def copy(self):

        result = CIMClass(self.classname)
        result.properties = self.properties.copy()
        result.methods = self.methods.copy()
        result.superclass = self.superclass
        result.qualifiers = self.qualifiers.copy()

        return result

    def __repr__(self):
        return "%s(%s, ...)" % (self.__class__.__name__, `self.classname`)

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, CIMClass):
            return 1

        return (cmpname(self.classname, other.classname)
                or cmpname(self.superclass, other.superclass)
                or cmp(self.properties, other.properties)
                or cmp(self.qualifiers, other.qualifiers)
                or cmp(self.methods, other.methods))
    
    def tocimxml(self):
        return cim_xml.CLASS(
            self.classname,
            properties = [p.tocimxml() for p in self.properties.values()],
            methods = [m.tocimxml() for m in self.methods.values()],
            qualifiers = [q.tocimxml() for q in self.qualifiers.values()],
            superclass = self.superclass)

    def tomof(self):

        def _makequalifiers(qualifiers, indent):
            """Return a mof fragment for a NocaseDict of qualifiers."""

            if len(qualifiers) == 0:
                return ''
            
            return '[%s]' % ',\n '.ljust(indent+2).join([q.tomof() for q in qualifiers.values()])

        # Class definition

        s = '   %s\n' % _makequalifiers(self.qualifiers, 4)

        s += 'class %s ' % self.classname

        # Superclass

        if self.superclass is not None:
            s += ': %s ' % self.superclass

        s += '{\n'

        # Properties

        for p in self.properties.values():
            s += '      %s\n' % (_makequalifiers(p.qualifiers, 7))
            s += '   %s %s;\n' % (p.type, p.name)

        # Methods

        for m in self.methods.values():
            s += '      %s\n' % (_makequalifiers(m.qualifiers, 7))
            s += '   %s\n' % m.tomof()

        s += '};\n'
        
        return s

class CIMMethod(object):

    def __init__(self, methodname, return_type = None, parameters = {}, 
                 class_origin = None, propagated = False, qualifiers = {}):

        self.name = methodname
        self.return_type = return_type
        self.parameters = NocaseDict(parameters)
        self.class_origin = class_origin
        self.propagated = propagated
        self.qualifiers = NocaseDict(qualifiers)

    def copy(self):

        result = CIMMethod(self.name,
                           return_type = self.return_type,
                           class_origin = self.class_origin,
                           propagated = self.propagated)

        result.parameters = self.parameters.copy()
        result.qualifiers = self.qualifiers.copy()

        return result

    def tocimxml(self):
        return cim_xml.METHOD(
            self.name,
            parameters = [p.tocimxml() for p in self.parameters.values()],
            return_type = self.return_type,
            class_origin = self.class_origin,
            propagated = self.propagated,
            qualifiers = [q.tocimxml() for q in self.qualifiers.values()])

    def __repr__(self):
        return '%s(name=%s, return_type=%s...)' % \
               (self.__class__.__name__, `self.name`, `self.return_type`)

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, CIMMethod):
            return 1

        return (cmpname(self.name, other.name) or
                cmp(self.parameters, other.parameters) or
                cmp(self.qualifiers, other.qualifiers) or
                cmp(self.class_origin, other.class_origin) or
                cmp(self.propagated, other.propagated) or
                cmp(self.return_type, other.return_type))

    def tomof(self):

        s = ''

        if self.return_type is not None:
            s += '%s ' % self.return_type

        s += '%s(%s);' % \
             (self.name,
              string.join([p.tomof() for p in self.parameters.values()], ', '))

        return s
    
class CIMParameter(object):

    def __init__(self, name, type, reference_class = None, is_array = None,
                 array_size = None, qualifiers = {}, value = None):

        self.name = name
        self.type = type
        self.reference_class = reference_class
        self.is_array = is_array
        self.array_size = array_size
        self.qualifiers = NocaseDict(qualifiers)
        self.value = value

    def copy(self):

        result = CIMParameter(self.name,
                              self.type,
                              reference_class = self.reference_class,
                              is_array = self.is_array,
                              array_size = self.array_size,
                              value = self.value)

        result.qualifiers = self.qualifiers.copy()

        return result

    def __repr__(self):

        return '%s(name=%s, type=%s, is_array=%s)' % \
               (self.__class__.__name__, `self.name`, `self.type`,
                `self.is_array`)

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, self.__class__):
            return 1

        return (cmpname(self.name, other.name) or
                cmp(self.type, other.type) or
                cmpname(self.reference_class, other.reference_class) or
                cmp(self.is_array, other.is_array) or
                cmp(self.array_size, other.array_size) or
                cmp(self.qualifiers, other.qualifiers) or
                cmp(self.value, other.value))

    def tocimxml(self):

        if self.type == 'reference':

            if self.is_array:

                array_size = None

                if self.array_size is not None:
                    array_size = str(self.array_size)

                return cim_xml.PARAMETER_REFARRAY(
                    self.name,
                    self.reference_class,
                    array_size,
                    qualifiers = [q.tocimxml()
                                  for q in self.qualifiers.values()])

            else:

                return cim_xml.PARAMETER_REFERENCE(
                    self.name,
                    self.reference_class,
                    qualifiers = [q.tocimxml()
                                  for q in self.qualifiers.values()])

        elif self.is_array:
            
            array_size = None

            if self.array_size is not None:
                array_size = str(self.array_size)

            return cim_xml.PARAMETER_ARRAY(
                self.name,
                self.type,
                array_size,
                qualifiers = [q.tocimxml() for q in self.qualifiers.values()])

        else:

            return cim_xml.PARAMETER(
                self.name,
                self.type,
                qualifiers = [q.tocimxml() for q in self.qualifiers.values()])

    def tomof(self):
        return '%s %s' % (self.type, self.name)

class CIMQualifier(object):
    """Represents static annotations of a class, method, property, etc.

    Includes information such as a documentation string and whether a property
    is a key."""
    
    def __init__(self, name, value, type = None, propagated = None,
                 overridable = None, tosubclass = None, toinstance = None,
                 translatable = None):

        self.name = name
        self.type = type
        self.propagated = propagated
        self.overridable = overridable
        self.tosubclass = tosubclass
        self.toinstance = toinstance
        self.translatable = translatable

        # Determine type of value if not specified

        import __builtin__
            
        if type is None:

            # Can't work out what is going on if type and value are
            # both not set.

            if value is None:
                raise TypeError('Null qualifier "%s" must have a type' % name)
        
            if __builtin__.type(value) == list:

                # Determine type for list value

                if len(value) == 0:
                    raise TypeError(
                        'Empty qualifier array "%s" must have a type' % name)
                    
                self.type = cim_types.cimtype(value[0])
                
            else:

                # Determine type for regular value
                
                self.type = cim_types.cimtype(value)

        # Don't let anyone set integer or float values.  You must use
        # a subclass from the cim_type module.

        if __builtin__.type(value) in (int, float, long):
            raise TypeError('Must use a CIM type for numeric qualifiers.')

        self.value = value

    def copy(self):

        return CIMQualifier(self.name,
                            self.value,
                            type = self.type,
                            propagated = self.propagated,
                            overridable = self.overridable,
                            tosubclass = self.tosubclass,
                            toinstance = self.toinstance,
                            translatable = self.translatable)
        
    def __repr__(self):
        return "%s(%s, %s)" % \
               (self.__class__.__name__, `self.name`, `self.value`)

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, CIMQualifier):
            return 1

        return (cmpname(self.name, other.name) or
                cmp(self.value, other.value) or
                cmp(self.type, other.type) or
                cmp(self.propagated, other.propagated) or
                cmp(self.overridable, other.overridable) or
                cmp(self.tosubclass, other.tosubclass) or
                cmp(self.toinstance, other.toinstance) or
                cmp(self.translatable, other.translatable))

    def tocimxml(self):

        value = None

        if type(self.value) == list:
            value = VALUE_ARRAY([VALUE(v) for v in self.value])
        elif self.value is not None:
            value = VALUE(self.value)

        return QUALIFIER(self.name,
                         self.type,
                         value,
                         propagated = self.propagated,
                         overridable = self.overridable,
                         tosubclass = self.tosubclass,
                         toinstance = self.toinstance,
                         translatable = self.translatable)
    
    def tomof(self):

        def valstr(v):
            if isinstance(v, basestring):
                return '"%s"' % v
            return str(v)

        if type(self.value) == list:
            return '%s {' % self.name + \
                   ', '.join([valstr(v) for v in self.value]) + '}'

        return '%s (%s)' % (self.name, valstr(self.value))

class CIMQualifierDeclaration(object):
    """Represents the declaration of a qualifier."""

    # TODO: Scope and qualifier flavors
    
    def __init__(self, name, type, value = None, is_array = False,
                 array_size = None, scopes = {}, 
                 overridable = None, tosubclass = None, toinstance = None,
                 translatable = None):

        self.name = name
        self.type = type
        self.value = value
        self.is_array = is_array
        self.array_size = array_size
        self.scopes = NocaseDict(scopes)
        self.overridable = overridable
        self.tosubclass = tosubclass
        self.toinstance = toinstance
        self.translatable = translatable

    def copy(self):

        return CIMQualifierDeclaration(self.name,
                                       self.type,
                                       value=self.value,
                                       is_array=self.is_array,
                                       array_size=self.array_size,
                                       scopes=self.scopes,
                                       overridable=self.overridable,
                                       tosubclass=self.tosubclass,
                                       toinstance=self.toinstance,
                                       translatable=self.translatable)
                                       
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, `self.name`)

    def __cmp__(self, other):

        if self is other:
            return 0
        elif not isinstance(other, CIMQualifierDeclaration):
            return 1

        return (cmpname(self.name, other.name) or
                cmp(self.type, other.type) or
                cmp(self.value, other.value) or
                cmp(self.is_array, other.is_array) or
                cmp(self.array_size, other.array_size) or
                cmp(self.scopes, other.scopes) or
                cmp(self.overridable, other.overridable) or
                cmp(self.tosubclass, other.tosubclass) or
                cmp(self.toinstance, other.toinstance) or
                cmp(self.translatable, other.translatable))

    def tocimxml(self):
        
        return QUALIFIER_DECLARATION(self.name,
                                     self.type,
                                     self.value,
                                     is_array = self.is_array,
                                     array_size = self.array_size,
                                     qualifier_scopes = self.scopes,
                                     overridable=self.overridable,
                                     tosubclass=self.tosubclass, 
                                     toinstance=self.toinstance,
                                     translatable=self.translatable)

    def tomof(self):
        mof = 'Qualifier %s : %s' % (self.name, self.type)
        if self.is_array:
            mof+= '['
            if self.array_size is not None:
                mof+= str(self.array_size)
            mof+= ']'
        if self.value is not None:
            if isinstance(self.value, list):
                mof+= '{'
                mof+= ', '.join([atomic_to_cim_xml(tocimobj(self.type, x)) \
                        for x in self.value])
                mof+= '}'
            else:
                mof+= ' = %s' % atomic_to_cim_xml(tocimobj(self.type,self.value))
        mof+= ',\n    '
        mof+= 'Scope('
        mof+= ', '.join([x.lower() for x, y in self.scopes.items() if y]) + ')'
        if not self.overridable and not self.tosubclass \
                and not self.toinstance and not self.translatable:
            mof+= ';'
            return mof
        mof+= ',\n    Flavor('
        mof+= self.overridable and 'EnableOverride' or 'DisableOverride'
        mof+= ', '
        mof+= self.tosubclass and 'ToSubclass' or 'Restricted'
        if self.toinstance:
            mof+= ', ToInstance'
        if self.translatable:
            mof+= ', Translatable'
        mof+= ');'
        return mof

def tocimxml(value):
    """Convert an arbitrary object to CIM xml.  Works with cim_obj
    objects and builtin types."""

    # Python cim_obj object

    if hasattr(value, 'tocimxml'):
        return value.tocimxml()

    # CIMType or builtin type

    if isinstance(value, cim_types.CIMType) or \
           type(value) in (str, unicode, int):
        return cim_xml.VALUE(unicode(value))

    if isinstance(value, bool):
        if value:
            return cim_xml.VALUE('TRUE')
        else:
            return cim_xml.VALUE('FALSE')
        raise TypeError('Invalid boolean type: %s' % value)

    # List of values

    if type(value) == list:
        return cim_xml.VALUE_ARRAY(map(tocimxml, value))

    raise ValueError("Can't convert %s (%s) to CIM XML" %
                     (`value`, type(value)))


def tocimobj(_type, value):
    """Convert a CIM type and a string value into an appropriate
    builtin type."""

    if value is None or _type is None:
        return None

    if _type != 'string' and isinstance(value, basestring) and not value:
        return None

    # Lists of values

    if type(value) == list:
        return map(lambda x: tocimobj(_type, x), value)

    # Boolean type
    
    if _type == 'boolean':
        if isinstance(value, bool):
            return value
        elif isinstance(value, basestring):
            if value.lower() == 'true':
                return True
            elif value.lower() == 'false':
                return False
        raise ValueError('Invalid boolean value "%s"' % value)

    # String type

    if _type == 'string':
        return value

    # Integer types

    if _type == 'uint8':
        return cim_types.Uint8(value)

    if _type == 'sint8':
        return cim_types.Sint8(value)

    if _type == 'uint16':
        return cim_types.Uint16(value)

    if _type == 'sint16':
        return cim_types.Sint16(value)

    if _type == 'uint32':
        return cim_types.Uint32(value)

    if _type == 'sint32':
        return cim_types.Sint32(value)

    if _type == 'uint64':
        return cim_types.Uint64(value)

    if _type == 'sint64':
        return cim_types.Sint64(value)

    # Real types

    if _type == 'real32':
        return cim_types.Real32(value)

    if _type == 'real64':
        return cim_types.Real64(value)

    # Char16

    if _type == 'char16':
        raise ValueError('CIMType char16 not handled')

    # Datetime

    if _type == 'datetime':
        return CIMDateTime(value)

    # REF
    def partition(s, seq):
        """ S.partition(sep) -> (head, sep, tail)

        Searches for the separator sep in S, and returns the part before it,
        the separator itself, and the part after it.  If the separator is not
        found, returns S and two empty strings.
        """
        try:
            return s.partition(seq)
        except AttributeError:
            try:
                idx = s.index(seq)
            except ValueError:
                return (s, '', '')
            return (s[:idx], seq, s[idx+len(seq):])

    if _type == 'reference':
        # TODO doesn't handle double-quoting, as in refs to refs.  Example:
        # r'ex_composedof.composer="ex_sampleClass.label1=9921,label2=\"SampleLabel\"",component="ex_sampleClass.label1=0121,label2=\"Component\""')
        if isinstance(value, (CIMInstanceName, CIMClassName)):
            return value
        elif isinstance(value, basestring):
            ns = host = None
            head, sep, tail = partition(value, '//')
            if sep and head.find('"') == -1:
                # we have a namespace type
                head, sep, tail = partition(tail, '/')
                host = head
            else:
                tail = head
            head, sep, tail = partition(tail, ':')
            if sep:
                ns = head
            else:
                tail = head
            head, sep, tail = partition(tail, '.')
            if not sep:
                return CIMClassName(head, host=host, namespace=ns)
            classname = head
            kb = {}
            while tail:
                head, sep, tail = partition(tail, ',')
                if head.count('"') == 1: # quoted string contains comma
                    tmp, sep, tail = partition(tail,'"')
                    head = '%s,%s' % (head, tmp)
                    tail = partition(tail,',')[2]
                head = head.strip()
                key, sep, val = partition(head,'=')
                if sep:
                    cn, s, k = partition(key, '.')
                    if s:
                        if cn != classname:
                            raise ValueError('Invalid object path: "%s"' % \
                                    value)
                        key = k
                    val = val.strip()
                    if val[0] == '"' and val[-1] == '"':
                        val = val.strip('"')
                    else:
                        if val.lower() in ('true','false'):
                            val = val.lower() == 'true'
                        elif val.isdigit():
                            val = int(val)
                        else:
                            try:
                                val = float(val)
                            except ValueError:
                                try:
                                    val = CIMDateTime(val)
                                except ValueError:
                                    raise ValueError('Invalid key binding: %s'\
                                            % val)
                                

                    kb[key] = val
            return CIMInstanceName(classname, host=host, namespace=ns, 
                    keybindings=kb)
        else:
            raise ValueError('Invalid reference value')

    raise ValueError('Invalid CIM type "%s"' % _type)


def byname(nlist):
    """Convert a list of named objects into a map indexed by name"""
    return dict([(x.name, x) for x in nlist])



#
# (C) Copyright 2003-2007 Hewlett-Packard Development Company, L.P.
# (C) Copyright 2006-2007 Novell, Inc. 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Tim Potter <tpot@hp.com>
#         Martin Pool <mbp@hp.com>
#         Bart Whiteley <bwhiteley@suse.de>

# This is meant to be safe for import *; ie the only global names
# should be ones that all clients can see.

import sys, string
from types import StringTypes
from xml.dom import minidom
import cim_obj, cim_xml, cim_http, cim_types
from cim_obj import CIMClassName, CIMInstanceName, CIMInstance, CIMClass
from datetime import datetime, timedelta
from tupletree import dom_to_tupletree, xml_to_tupletree
from tupleparse import parse_cim

"""CIM-XML/HTTP operations.

The WBEMConnection class opens a connection to a remote WBEM server.
Across this you can run various CIM operations.  Each method of this
object corresponds fairly directly to a single CIM method call.
"""

DEFAULT_NAMESPACE = 'root/cimv2'

# TODO: Many methods have more parameters that aren't set yet.

# helper functions for validating arguments

def _check_classname(val):
    if not isinstance(val, StringTypes):
        raise ValueError("string expected for classname, not %s" % `val`)


class CIMError(Exception):
    """Raised when something bad happens.  The associated value is a
    tuple of (error_code, description).  An error code of zero
    indicates an XML parsing error in PyWBEM."""

class WBEMConnection(object):
    """Class representing a client's connection to a WBEM server.
    
    At the moment there is no persistent TCP connection; the
    connectedness is only conceptual.

    After creating a connection, various methods may be called on the
    object, which causes a remote call to the server.  All these
    operations take regular Python or cim_types values for parameters,
    and return the same.  The caller should not need to know about
    the XML encoding.  (It should be possible to use a different
    transport below this layer without disturbing any clients.)

    The connection remembers the XML for the last request and last
    reply.  This may be useful in debugging: if a problem occurs, you
    can examine the last_request and last_reply fields of the
    connection.  These are the prettified request and response; the
    real request is sent without indents so as not to corrupt whitespace.

    The caller may also register callback functions which are passed
    the request before it is sent, and the reply before it is
    unpacked.

    verify_callback is used to verify the server certificate.  
    It is passed to OpenSSL.SSL.set_verify, and is called during the SSL
    handshake.  verify_callback should take five arguments: A Connection 
    object, an X509 object, and three integer variables, which are in turn 
    potential error number, error depth and return code. verify_callback 
    should return True if verification passes and False otherwise.

    The value of the x509 argument is used only when the url contains
    'https'. x509 must be a dictionary containing the keys 'cert_file' 
    and 'key_file'. The value of 'cert_file' must consist of the
    filename of an certificate and the value of 'key_file' must consist 
    of a filename containing the private key belonging to the public key 
    that is part of the certificate in cert_file. 
    """
    
    def __init__(self, url, creds = None, default_namespace = DEFAULT_NAMESPACE,
                 x509 = None, verify_callback = None):
        self.url = url
        self.creds = creds
        self.x509 = x509
        self.verify_callback = verify_callback
        self.last_request = self.last_reply = ''
        self.default_namespace = default_namespace
        self.debug = False

    def __repr__(self):
        if self.creds is None:
            user = 'anonymous'
        else:
            user = 'user=%s' % `self.creds[0]`
        return "%s(%s, %s, namespace=%s)" % (self.__class__.__name__, `self.url`,
                                             user, `self.default_namespace`)

    def imethodcall(self, methodname, namespace, **params):
        """Make an intrinsic method call.

        Returns a tupletree with a IRETURNVALUE element at the root.
        A CIMError exception is thrown if there was an error parsing
        the call response, or an ERROR element was returned.

        The parameters are automatically converted to the right
        CIM_XML objects.

        In general clients should call one of the method-specific
        methods of the connection, such as EnumerateInstanceNames,
        etc."""

        # Create HTTP headers

        headers = ['CIMOperation: MethodCall',
                   'CIMMethod: %s' % methodname,
                   cim_http.get_object_header(namespace)]

        # Create parameter list

        plist = map(lambda x:
                    cim_xml.IPARAMVALUE(x[0], cim_obj.tocimxml(x[1])),
                    params.items())
        
        # Build XML request

        req_xml = cim_xml.CIM(
            cim_xml.MESSAGE(
                cim_xml.SIMPLEREQ(
                    cim_xml.IMETHODCALL(
                        methodname,
                        cim_xml.LOCALNAMESPACEPATH(
                            [cim_xml.NAMESPACE(ns)
                             for ns in string.split(namespace, '/')]),
                        plist)),
                '1001', '1.0'),
            '2.0', '2.0')

        if self.debug:
            self.last_raw_request = req_xml.toxml()
            self.last_request = req_xml.toprettyxml(indent='  ')

            self.last_reply = None
            self.last_raw_reply = None
        
        # Get XML response

        try:
            resp_xml = cim_http.wbem_request(self.url, req_xml.toxml(),
                                             self.creds, headers,
                                             x509 = self.x509,
                                             verify_callback = self.verify_callback)
        except cim_http.AuthError:
            raise
        except cim_http.Error, arg:
            # Convert cim_http exceptions to CIMError exceptions
            raise CIMError(0, str(arg))

        ## TODO: Perhaps only compute this if it's required?  Should not be
        ## all that expensive.

        reply_dom = minidom.parseString(resp_xml)

        if self.debug:
            self.last_reply = reply_dom.toprettyxml(indent='  ')
            self.last_raw_reply = resp_xml

        # Parse response

        tt = parse_cim(dom_to_tupletree(reply_dom))

        if tt[0] != 'CIM':
            raise CIMError(0, 'Expecting CIM element, got %s' % tt[0])
        tt = tt[2]
        
        if tt[0] != 'MESSAGE':
            raise CIMError(0, 'Expecting MESSAGE element, got %s' % tt[0])
        tt = tt[2]

        if len(tt) != 1 or tt[0][0] != 'SIMPLERSP':
            raise CIMError(0, 'Expecting one SIMPLERSP element')
        tt = tt[0][2]
        
        if tt[0] != 'IMETHODRESPONSE':
            raise CIMError(
                0, 'Expecting IMETHODRESPONSE element, got %s' % tt[0])

        if tt[1]['NAME'] != methodname:
            raise CIMError(0, 'Expecting attribute NAME=%s, got %s' %
                           (methodname, tt[1]['NAME']))
        tt = tt[2]

        # At this point we either have a IRETURNVALUE, ERROR element
        # or None if there was no child nodes of the IMETHODRESPONSE
        # element.

        if tt is None:
            return None

        if tt[0] == 'ERROR':
            code = int(tt[1]['CODE'])
            if tt[1].has_key('DESCRIPTION'):
                raise CIMError(code, tt[1]['DESCRIPTION'])
            raise CIMError(code, 'Error code %s' % tt[1]['CODE'])

        if tt[0] != 'IRETURNVALUE':
            raise CIMError(0, 'Expecting IRETURNVALUE element, got %s' % tt[0])

        return tt

    def methodcall(self, methodname, localobject, **params):
        """Make an extrinsic method call.

        Returns a tupletree with a RETURNVALUE element at the root.
        A CIMError exception is thrown if there was an error parsing
        the call response, or an ERROR element was returned.

        The parameters are automatically converted to the right
        CIM_XML objects."""

        # METHODCALL only takes a LOCALCLASSPATH or LOCALINSTANCEPATH 
        if hasattr(localobject, 'host') and localobject.host is not None:
            localobject = localobject.copy()
            localobject.host = None

        # Create HTTP headers

        headers = ['CIMOperation: MethodCall',
                   'CIMMethod: %s' % methodname,
                   cim_http.get_object_header(localobject)]
            
        # Create parameter list

        def paramtype(obj):
            """Return a string to be used as the CIMTYPE for a parameter."""
            if isinstance(obj, cim_types.CIMType):
                return obj.cimtype
            elif type(obj) == bool:
                return 'boolean'
            elif isinstance(obj, StringTypes):
                return 'string'
            elif isinstance(obj, (datetime, timedelta)):
                return 'datetime'
            elif isinstance(obj, (CIMClassName, CIMInstanceName)):
                return 'reference'
            elif isinstance(obj, (CIMClass, CIMInstance)):
                return 'string'
            elif isinstance(obj, list):
                if obj:
                    return paramtype(obj[0])
                else:
                    return None
            raise TypeError('Unsupported parameter type "%s"' % type(obj))

        def paramvalue(obj):
            """Return a cim_xml node to be used as the value for a
            parameter."""
            if isinstance(obj, (datetime, timedelta)):
                obj = cim_types.CIMDateTime(obj)
            if isinstance(obj, (cim_types.CIMType, bool, StringTypes)):
                return cim_xml.VALUE(cim_types.atomic_to_cim_xml(obj))
            if isinstance(obj, (CIMClassName, CIMInstanceName)):
                return cim_xml.VALUE_REFERENCE(obj.tocimxml())
            if isinstance(obj, (CIMClass, CIMInstance)):
                return cim_xml.VALUE(obj.tocimxml().toxml())
            if isinstance(obj, list):
                if obj and isinstance(obj[0], (CIMClassName, CIMInstanceName)):
                    return cim_xml.VALUE_REFARRAY([paramvalue(x) for x in obj])
                return cim_xml.VALUE_ARRAY([paramvalue(x) for x in obj])
            raise TypeError('Unsupported parameter type "%s"' % type(obj))

        def is_embedded(obj):
            """Determine if an object requires an EmbeddedObject attribute"""
            if isinstance(obj,list) and obj:
                return is_embedded(obj[0])
            elif isinstance(obj, CIMClass):
                return 'object'
            elif isinstance(obj, CIMInstance):
                return 'instance'
            return None

        plist = [cim_xml.PARAMVALUE(x[0], 
                                    paramvalue(x[1]), 
                                    paramtype(x[1]),
                                    embedded_object=is_embedded(x[1]))
                 for x in params.items()]

        # Build XML request

        req_xml = cim_xml.CIM(
            cim_xml.MESSAGE(
                cim_xml.SIMPLEREQ(
                    cim_xml.METHODCALL(
                        methodname,
                        localobject.tocimxml(),
                        plist)),
                '1001', '1.0'),
            '2.0', '2.0')

        if self.debug:
            self.last_request = req_xml.toprettyxml(indent='  ')

        # Get XML response

        try:
            resp_xml = cim_http.wbem_request(self.url, req_xml.toxml(),
                                             self.creds, headers,
                                             x509 = self.x509,
                                             verify_callback = self.verify_callback)
        except cim_http.Error, arg:
            # Convert cim_http exceptions to CIMError exceptions
            raise CIMError(0, str(arg))

        if self.debug:
            self.last_reply = resp_xml

        tt = parse_cim(xml_to_tupletree(resp_xml))

        if tt[0] != 'CIM':
            raise CIMError(0, 'Expecting CIM element, got %s' % tt[0])
        tt = tt[2]
        
        if tt[0] != 'MESSAGE':
            raise CIMError(0, 'Expecting MESSAGE element, got %s' % tt[0])
        tt = tt[2]

        if len(tt) != 1 or tt[0][0] != 'SIMPLERSP':
            raise CIMError(0, 'Expecting one SIMPLERSP element')
        tt = tt[0][2]
        
        if tt[0] != 'METHODRESPONSE':
            raise CIMError(
                0, 'Expecting METHODRESPONSE element, got %s' % tt[0])

        if tt[1]['NAME'] != methodname:
            raise CIMError(0, 'Expecting attribute NAME=%s, got %s' %
                           (methodname, tt[1]['NAME']))
        tt = tt[2]

        # At this point we have an optional RETURNVALUE and zero or
        # more PARAMVALUE elements representing output parameters.

        if len(tt) > 0 and tt[0][0] == 'ERROR':
            code = int(tt[0][1]['CODE'])
            if tt[0][1].has_key('DESCRIPTION'):
                raise CIMError(code, tt[0][1]['DESCRIPTION'])
            raise CIMError(code, 'Error code %s' % tt[0][1]['CODE'])

        return tt

    #
    # Instance provider API
    # 

    def EnumerateInstanceNames(self, ClassName, namespace = None, **params):
        """Enumerate instance names of a given classname.  Returns a
        list of CIMInstanceName objects."""
        
        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'EnumerateInstanceNames',
            namespace,
            ClassName = CIMClassName(ClassName),
            **params)

        names = []

        if result is not None:
            names = result[2]

        [setattr(n, 'namespace', namespace) for n in names]

        return names

    def EnumerateInstances(self, ClassName, namespace = None, **params):
        """Enumerate instances of a given classname.  Returns a list
        of CIMInstance objects with paths."""

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'EnumerateInstances',
            namespace,
            ClassName = CIMClassName(ClassName),
            **params)

        instances = []

        if result is not None:
            instances = result[2]

        [setattr(i.path, 'namespace', namespace) for i in instances]

        return instances

    def GetInstance(self, InstanceName, **params):
        """Fetch an instance given by instancename.  Returns a
        CIMInstance object."""

        # Strip off host and namespace to make this a "local" object
        
        iname = InstanceName.copy()
        iname.host = None
        iname.namespace = None

        if InstanceName.namespace is None:
            namespace = self.default_namespace
        else:
            namespace = InstanceName.namespace

        result = self.imethodcall(
            'GetInstance',
            namespace,
            InstanceName = iname,
            **params)

        instance = result[2][0]
        instance.path = InstanceName
        instance.path.namespace = namespace

        return instance

    def DeleteInstance(self, InstanceName, **params):
        """Delete the instance given by instancename."""

        # Strip off host and namespace to make this a "local" object

        iname = InstanceName.copy()
        iname.host = None
        iname.namespace = None        

        if InstanceName.namespace is None:
            namespace = self.default_namespace
        else:
            namespace = InstanceName.namespace

        self.imethodcall(
            'DeleteInstance',
            namespace,
            InstanceName = iname,
            **params)

    def CreateInstance(self, NewInstance, **params):
        """Create an instance.  Returns the name for the instance."""

        # Take namespace path from object parameter

        if NewInstance.path is not None and \
           NewInstance.path.namespace is not None:
            namespace = NewInstance.path.namespace
        else:
            namespace = self.default_namespace

        # Strip off path to avoid producing a VALUE.NAMEDINSTANCE
        # element instead of an INSTANCE element.

        instance = NewInstance.copy()
        instance.path = None

        result = self.imethodcall(
            'CreateInstance',
            namespace,
            NewInstance = instance,
            **params)

        name = result[2][0]
        name.namespace = namespace

        return name

    def ModifyInstance(self, ModifiedInstance, **params):
        """Modify properties of a named instance."""

        # Must pass a named CIMInstance here (i.e path attribute set)

        if ModifiedInstance.path is None:
            raise ValueError(
                'ModifiedInstance parameter must have path attribute set')

        # Take namespace path from object parameter

        if ModifiedInstance.path.namespace is None:
            namespace = self.default_namespace
        else:
            namespace = ModifiedInstance.path.namespace        

        instance = ModifiedInstance.copy()
        instance.path.namespace = None

        self.imethodcall(
            'ModifyInstance',
            namespace,
            ModifiedInstance = instance,
            **params)

    def ExecQuery(self, QueryLanguage, Query, namespace = None):
        if namespace is None:
            namespace = self.default_namespace
        result = self.imethodcall(
            'ExecQuery',
            namespace, 
            QueryLanguage = QueryLanguage,
            Query = Query)

        instances = []

        if result is not None:
            instances = [tt[2] for tt in result[2]]

        [setattr(i.path, 'namespace', namespace) for i in instances]

        return instances

    #
    # Schema management API
    #

    def _map_classname_param(self, params):
        """Convert string ClassName parameter to a CIMClassName."""

        if params.has_key('ClassName') and \
           isinstance(params['ClassName'], StringTypes):
            params['ClassName'] = cim_obj.CIMClassName(params['ClassName'])

        return params

    def EnumerateClassNames(self, namespace = None, **params):
        """Return a list of CIM class names. Names are returned as strings."""
        
        params = self._map_classname_param(params)

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'EnumerateClassNames',
            namespace,
            **params)

        if result is None:
            return []
        else:
            return map(lambda x: x.classname, result[2])
        
    def EnumerateClasses(self, namespace = None, **params):
        """Return a list of CIM class objects."""

        params = self._map_classname_param(params)

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'EnumerateClasses',
            namespace,
            **params)

        if result is None:
            return []
        
        return result[2]

    def GetClass(self, ClassName, namespace = None, **params):
        """Return a CIMClass representing the named class."""

        params = self._map_classname_param(params)

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'GetClass',
            namespace,
            ClassName = CIMClassName(ClassName),
            **params)

        return result[2][0]

    def DeleteClass(self, ClassName, namespace = None, **params):
        """Delete a class by class name."""

        # UNSUPPORTED (but actually works)

        params = self._map_classname_param(params)

        if namespace is None:
            namespace = self.default_namespace

        self.imethodcall(
            'DeleteClass',
            namespace,
            ClassName = CIMClassName(ClassName),
            **params)            

    def ModifyClass(self, ModifiedClass, namespace = None, **params):
        """Modify a CIM class."""

        # UNSUPPORTED

        if namespace is None:
            namespace = self.default_namespace

        self.imethodcall(
            'ModifyClass',
            namespace,
            ModifiedClass = ModifiedClass,
            **params)

    def CreateClass(self, NewClass, namespace = None, **params):
        """Create a CIM class."""

        # UNSUPPORTED

        if namespace is None:
            namespace = self.default_namespace

        self.imethodcall(
            'CreateClass',
            namespace,
            NewClass = NewClass,
            **params)

    #
    # Association provider API
    # 

    def _add_objectname_param(self, params, object):
        """Add an object name (either a class name or an instance
        name) to a dictionary of parameter names."""

        if isinstance(object, (CIMClassName, CIMInstanceName)):
            params['ObjectName'] = object.copy()
            params['ObjectName'].namespace = None
        elif isinstance(object, StringTypes):
            params['ObjectName'] = CIMClassName(object)
        else:
            raise ValueError('Expecting a classname, CIMClassName or '
                             'CIMInstanceName object')

        return params

    def _map_association_params(self, params):
        """Convert various convenience parameters and types into their
        correct form for passing to the imethodcall() function."""

        # ResultClass and Role parameters that are strings should be
        # mapped to CIMClassName objects.

        if params.has_key('ResultClass') and \
           isinstance(params['ResultClass'], StringTypes):
            params['ResultClass'] = cim_obj.CIMClassName(params['ResultClass'])

        if params.has_key('AssocClass') and \
           isinstance(params['AssocClass'], StringTypes):
            params['AssocClass'] = cim_obj.CIMClassName(params['AssocClass'])

        return params

    def Associators(self, ObjectName, **params):
        """Enumerate CIM classes or instances that are associated to a
        particular source CIM Object.  Pass a keyword parameter of
        'ClassName' to return associators for a CIM class, pass
        'InstanceName' to return the associators for a CIM instance."""

        params = self._map_association_params(params)
        params = self._add_objectname_param(params, ObjectName)
        
        namespace = self.default_namespace
        
        if isinstance(ObjectName, CIMInstanceName) and \
           ObjectName.namespace is not None:
            namespace = ObjectName.namespace

        result = self.imethodcall(
            'Associators',
            namespace,
            **params)

        if result is None:
            return []

        return map(lambda x: x[2], result[2])

    def AssociatorNames(self, ObjectName, **params):
        """Enumerate the names of CIM classes or instances names that are
        associated to a particular source CIM Object.  Pass a keyword
        parameter of 'ClassName' to return associator names for a CIM
        class, pass 'InstanceName' to return the associator names for a CIM
        instance.  Returns a list of CIMInstanceName objects with the
        host and namespace attributes set."""

        params = self._map_association_params(params)
        params = self._add_objectname_param(params, ObjectName)

        namespace = self.default_namespace
        
        if isinstance(ObjectName, CIMInstanceName) and \
           ObjectName.namespace is not None:
            namespace = ObjectName.namespace

        result = self.imethodcall(
            'AssociatorNames',
            namespace,
            **params)

        if result is None:
            return []

        return map(lambda x: x[2], result[2])

    def References(self, ObjectName, **params):
        """Enumerate the association objects that refer to a
        particular target CIM class or instance.  Pass a keyword
        parameter of 'ClassName' to return associators for a CIM
        class, pass 'InstanceName' to return the associators for a CIM
        instance."""

        params = self._map_association_params(params)
        params = self._add_objectname_param(params, ObjectName)

        namespace = self.default_namespace
        
        if isinstance(ObjectName, CIMInstanceName) and \
           ObjectName.namespace is not None:
            namespace = ObjectName.namespace

        result = self.imethodcall(
            'References',
            namespace,
            **params)

        if result is None:
            return []

        return map(lambda x: x[2], result[2])

    def ReferenceNames(self, ObjectName, **params):
        """Enumerate the name of association objects that refer to a
        particular target CIM class or instance.  Pass a keyword
        parameter of 'ClassName' to return associators for a CIM
        class, pass 'InstanceName' to return the associators for a CIM
        instance."""

        params = self._map_association_params(params)
        params = self._add_objectname_param(params, ObjectName)
 
        namespace = self.default_namespace
        
        if isinstance(ObjectName, CIMInstanceName) and \
           ObjectName.namespace is not None:
            namespace = ObjectName.namespace

        result = self.imethodcall(
            'ReferenceNames',
            namespace,
            **params)

        if result is None:
            return []

        return map(lambda x: x[2], result[2])

    #
    # Method provider API
    #

    def InvokeMethod(self, MethodName, ObjectName, **params):

        # Convert string to CIMClassName

        obj = ObjectName

        if isinstance(obj, StringTypes):
            obj = CIMClassName(obj, namespace = self.default_namespace)

        if isinstance(obj, CIMInstanceName) and obj.namespace is None:
            obj = ObjectName.copy()
            obj.namespace = self.default_namespace

        # Make the method call

        result = self.methodcall(MethodName, obj, **params)

        # Convert optional RETURNVALUE into a Python object

        returnvalue = None

        if len(result) > 0 and result[0][0] == 'RETURNVALUE':

            returnvalue = cim_obj.tocimobj(result[0][1]['PARAMTYPE'],
                                           result[0][2])
            result = result[1:]

        # Convert zero or more PARAMVALUE elements into dictionary

        output_params = {}

        for p in result:
            if p[1] == 'reference':
                output_params[p[0]] = p[2]
            else:
                output_params[p[0]] = cim_obj.tocimobj(p[1], p[2])

        return returnvalue, output_params

    #
    # Qualifiers API
    #

    def EnumerateQualifiers(self, namespace = None, **params):
        """Enumerate qualifier declarations.  Returns a list of
        CIMQualifier objects."""

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'EnumerateQualifiers',
            namespace,
            **params)

        qualifiers = []

        if result is not None:
            names = result[2]
        else:
            names = []

        return names

    def GetQualifier(self, QualifierName, namespace = None, **params):
        """Retrieve a qualifier by name.  Returns a CIMQualifier
        object."""

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'GetQualifier',
            namespace,
            QualifierName = QualifierName,
            **params)

        if result is not None:
            names = result[2][0]

        return names

    def SetQualifier(self, QualifierDeclaration, namespace = None,
                     **params):
        """Set a qualifier."""

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'SetQualifier',
            namespace,
            QualifierDeclaration = QualifierDeclaration,
            **params)

    def DeleteQualifier(self, QualifierName, namespace = None,
                        **params):
        """Delete a qualifier by name."""

        if namespace is None:
            namespace = self.default_namespace

        result = self.imethodcall(
            'DeleteQualifier',
            namespace,
            QualifierName = QualifierName,
            **params)

def is_subclass(ch, ns, super, sub):
    """Determine if one class is a subclass of another

    Keyword Arguments:
    ch -- A CIMOMHandle.  Either a pycimmb.CIMOMHandle or a 
        pywbem.WBEMConnection.
    ns -- Namespace.
    super -- A string containing the super class name.
    sub -- The subclass.  This can either be a string or a pywbem.CIMClass.

    """

    lsuper = super.lower()
    if isinstance(sub, CIMClass):
        subname = sub.classname
        subclass = sub
    else:
        subname = sub
        subclass = None
    if subname.lower() == lsuper:
        return True
    if subclass is None:
        subclass = ch.GetClass(subname,
                               ns,
                               LocalOnly=True, 
                               IncludeQualifiers=False,
                               PropertyList=[],
                               IncludeClassOrigin=False)
    while subclass.superclass is not None:
        if subclass.superclass.lower() == lsuper:
            return True
        subclass = ch.GetClass(subclass.superclass,
                               ns,
                               LocalOnly=True, 
                               IncludeQualifiers=False,
                               PropertyList=[],
                               IncludeClassOrigin=False)
    return False

def PegasusUDSConnection(creds = None, **kwargs):
    return WBEMConnection('/var/run/tog-pegasus/cimxml.socket', creds, **kwargs)

def SFCBUDSConnection(creds = None, **kwargs):
    return WBEMConnection('/tmp/sfcbHttpSocket', creds, **kwargs)

def OpenWBEMUDSConnection(creds = None, **kwargs):
    return WBEMConnection('/tmp/OW@LCL@APIIPC_72859_Xq47Bf_P9r761-5_J-7_Q', 
                          creds, **kwargs)
